"Base hierarchical cross-Multiscale Entropy function."
import numpy as np
from matplotlib.pyplot import figure, show, subplot
import matplotlib.gridspec as gridspec
from copy import deepcopy

def hXMSEn(Sig1, Sig2, Mbjx, Scales=3, RadNew=0, Plotx=False):
    """hXMSEn  returns the hierarchical cross-entropy between two univariate data sequences.

    .. code-block:: python
    
        MSx, Sn, CI = hXMSEn(Sig1, Sig2, Mobj) 
        
    Returns a vector of cross-entropy values (``MSx``) calculated at each node 
    in the hierarchical tree, the average cross-entropy value across all 
    nodes at each scale (``Sn``), and the complexity index (``CI``) of the hierarchical 
    tree (i.e. ``sum(Sn)``) between the data sequences contained in ``Sig1`` and``Sig2`` using
    the parameters specified by the multiscale object (``Mobj``) over 3 temporal scales (default).
    The entropy values in ``MSx`` are ordered from the root node (S_00) to the
    Nth subnode at scale T (S_TN): i.e. S_00, S_10, S_11, S_20, S_21, S_22,
    S_23, S_30, S_31, S_32, S_33, S_34, S_35, S_36, S_37, S_40, ... , S_TN.
    The average cross-entropy values in Sn are ordered in the same way, with the
    value of the root node given first: i.e. S0, S1, S2, ..., ST
     
    .. code-block:: python
    
        MSx, Sn, CI = hXMSEn(Sig1, Sig2, Mobj, Keyword = value, ...)
        
    Returns a vector of cross-entropy values (``MSx``) calculated at each node 
    in the hierarchical tree, the average cross-entropy value across all
    nodes at each scale (``Sn``), and the complexity index (``CI``) of the entire
    hierarchical tree between the data sequences contained in ``Sig1`` and ``Sig2`` using 
    the following keyword arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default: 3)
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the cross-entropy specified by ``Mobj`` is ``XSampEn`` or ``XApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                    
                        * [1] Pooled Standard Deviation   - ``r*std(Ykj)``
                        * [2] Pooled Variance             - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                        
        :Plotx:    - When ``Plotx == True``, returns a plot of the average cross-entropy value at each time scale (i.e. the multiscale cross-entropy curve) and a hierarchical graph showing the cross-entropy value of each node in the hierarchical tree decomposition.  (default: False)
    
    :See also:
        ``MSobject``, ``XMSEn``, ``rXMSEn``, ``cXMSEn``, ``XSampEn``, ``XApEn``, ``MSEn``, ``hMSEn``, ``rMSEn``, ``cMSEn``
    
    :References:
        [1]   Matthew W. Flood,
            "hXMSEn - EntropyHub Project"
            2021, https://github.com/MattWillFlood/EntropyHub
        [2]   Rui Yan, Zhuo Yang, and Tao Zhang,
            "Multiscale cross entropy: a novel algorithm for analyzing two
            time series." 
            5th International Conference on Natural Computation. 
            Vol. 1, pp: 411-413 IEEE, 2009.
    
        [3] Ying Jiang, C-K. Peng and Yuesheng Xu,
            "Hierarchical entropy analysis for biological signals."
            Journal of Computational and Applied Mathematics
            236.5 (2011): 728-742.
    
    """
    
    Mobj = deepcopy(Mbjx)    
    S1 = np.squeeze(Sig1)
    S2 = np.squeeze(Sig2)
    assert Sig1.ndim==1 and Sig2.ndim==1,   "Sig1/Sig2:   must be numpy vectors (N>10)" 
    N1 = S1.shape[0]
    N2 = S2.shape[0]
                   
    assert N1>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"      
    assert isinstance(Mobj,object) and Mobj.Func.__name__[0]=='X', "Mobj:  must \
    be a x-multiscale entropy object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert (np.isin(RadNew,np.arange(1,5)) and Mobj.Func.__name__ in \
    ['XSampEn','XApEn']) or RadNew==0, "RadNew:     must be an integer in range \
    [1 4] and entropy function must be 'XSampEn' or 'XApEn'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
              
    if Mobj.Func.__name__ in 'XSampEn': Mobj.Kwargs['Vcp'] = False 
    
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x,y: np.sqrt((np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1))
        elif RadNew == 2:
            Rnew = lambda x,y: (np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1)
        elif RadNew == 3:
            Rnew = lambda x,y: np.mean(abs(np.hstack((x,y))-np.mean(np.hstack((x,y)))))
        elif RadNew == 4:
            Rnew = lambda x,y: np.median(abs(np.hstack((x,y))-np.median(np.hstack((x,y)))))   
                
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Pooled Standard Deviation','Pooled Variance','Joint Mean Abs Deviation', 'Joint Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])              
            Cx = .2
        
    XX, YY, N1, N2 = Hierarchy(S1,S2,Scales)
    MSx = np.zeros(XX.shape[0])
    for T in range(XX.shape[0]):
        print(' .', end='')
        
        Xa = XX[T,:int(N1/(2**(int(np.log2(T+1)))))]
        Xb = YY[T,:int(N2/(2**(int(np.log2(T+1)))))]

        if RadNew:
            Mobj.Kwargs.update({'r': Cx*Rnew(Xa,Xb)})            
        Temp2 = Mobj.Func(Xa,Xb,**Mobj.Kwargs)
        
        if isinstance(Temp2,tuple):
            if isinstance(Temp2[0],(int,float)):
                MSx[T] = Temp2[0]
            else:
                MSx[T] =Temp2[0][-1]
        elif isinstance(Temp2,(int,float)):
            MSx[T] = Temp2
        elif isinstance(Temp2,np.ndarray):
            MSx[T] = Temp2[-1]
                    
    Sn = np.zeros(Scales)
    for t in range(Scales):
        Sn[t] = np.mean(MSx[(2**t)-1:(2**(t+1))-1])
       
    CI = sum(Sn)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')
    print('\n')
            
    if Plotx:       
       
        figure()
        G = gridspec.GridSpec(10, 1)    
        ax1 = subplot(G[:2, :])
        ax1.plot(np.arange(1,Scales+1), Sn, color=(8/255, 63/255, 77/255), linewidth=3)
        ax1.scatter(np.arange(1,Scales+1), Sn, 60, color=(1,0,1))
        ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_title('Hierarchical Multiscale %s Entropy'%(Mobj.Func.__name__), 
                      fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))         

        N = 2**(Scales-1)
        x = np.zeros(2*N - 1,dtype=int)  
        x[0] = N;        
        y = Scales*(Scales - np.log2(np.arange(1,2*N))//1)
        for k in range(1,2*N):
            Q = int(np.log2(k)//1)
            P = int((k)//2)-1            
            if k>1:              
                if k%2:
                    x[k-1] =  x[P] + N/(2**Q)
                else:
                    x[k-1] =  x[P] - N/(2**Q)
                    
        Edges = np.vstack((np.repeat(np.arange(1,N),2),np.arange(2,2*N))).transpose() - 1
        labx = ["".join(k) for k in np.round(MSx,3).astype(str)]
        ax2 = subplot(G[3:,:])   
        ax2.scatter(x,y,s=100*(MSx-min(MSx)+1)/(abs(min(MSx))+1),color=(1,0,1))
        for k in range(len(x)-1):            
            ax2.plot(x[Edges[k,:]],y[Edges[k,:]],color=(8/255,63/255,77/255),linewidth=2.5)
            ax2.annotate(labx[k],(x[k],y[k]))  
        ax2.annotate(labx[-1],(x[-1],y[-1]))
        show()
    
    return MSx, Sn, CI

    
def Hierarchy(Za, Zb, sx):
    Na = int(2**np.floor(np.log2(len(Za))))
    if np.log2(len(Za))%1 != 0:
        print('WARNING: Only first %d samples were used in hierarchical decomposition. \
            \nThe last %d samples of the data sequence were ignored.'%(Na,len(Za)-Na))
    if Na/(2**(sx-1)) < 8:
       raise Exception('Data length (%d) is too short to estimate entropy at the lowest' \
                      ' subtree. Consider reducing the number of scales.'%Na) 
                      
    Nb = int(2**np.floor(np.log2(len(Zb))))
    if np.log2(len(Zb))%1 != 0:
        print('WARNING: Only first %d samples were used in hierarchical decomposition. \
            \nThe last %d samples of the data sequence were ignored.'%(Nb,len(Zb)-Nb))
    if Nb/(2**(sx-1)) < 8:
       raise Exception('Data length (%d) is too short to estimate entropy at the lowest' \
                      ' subtree. Consider reducing the number of scales.'%Nb) 
    
    Za = Za[:Na]
    Zb = Zb[:Nb]
    
    U1 = np.zeros(((2**sx)-1,Na))
    U1[0,:] = Za
    p=1
    for k in range(sx-1):
        for n in range(2**k):
            Temp = U1[(2**k)+n-1,:]        
            U1[p,:Na//2]  = (Temp[::2] + Temp[1::2])/2
            U1[p+1,:Na//2]= (Temp[::2] - Temp[1::2])/2            
            p+=2
            
    U2 = np.zeros(((2**sx)-1,Nb))
    U2[0,:] = Zb
    p=1
    for k in range(sx-1):
        for n in range(2**k):
            Temp = U2[(2**k)+n-1,:]        
            U2[p,:Nb//2]  = (Temp[::2] + Temp[1::2])/2
            U2[p+1,:Nb//2]= (Temp[::2] - Temp[1::2])/2
            p+=2
                            
    return U1, U2, Na, Nb

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Dispersion Entropy function."""
import numpy as np       
from scipy.special import ndtr
from scipy.cluster.vq import kmeans2

def DispEn(Sig, m=2, tau=1, c=3, Typex='NCDF', Logx=np.exp(1), Fluct=False, Norm=False, rho=1):
    """DispEn  estimates the dispersion entropy of a univariate data sequence.

    .. code-block:: python
    
        Dispx, Ppi = DispEn(Sig)
    
    Returns the dispersion entropy (``Dispx``) and the reverse dispersion entropy
    (``RDE``) estimated from the data sequence (``Sig``) using the default parameters:
    embedding dimension = 2, time delay = 1, symbols = 3, logarithm = natural,
    data transform = normalised cumulative density function (ncdf)
 
    .. code-block:: python
        
        Dispx, Ppi = DispEn(Sig, keyword = value, ...)
        
    Returns the dispersion entropy (``Dispx``) and the reverse dispersion entropy (``RDE``)
    estimated from the data sequence (``Sig``) using the specified 'keyword' arguments:
        :m:     - Embedding Dimension,  a positive integer
        :tau:   - Time Delay, a positive integer
        :c:     - Number of symbols, an integer > 1
        :Typex: - Typex of data-to-symbolic sequence transform, one of the following strings: {``'linear'``, ``'kmeans'``, ``'ncdf'``, ``'finesort'``, ``'equal'``}
            See the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_ for more info on these transforms.
                    
        :Logx:  - Logarithm base, a positive scalar
        :Fluct: - When ``Fluct == True``, DispEn returns the fluctuation-based Dispersion entropy.   [default: False]
        :Norm:  - Normalisation of ``Dispx`` and ``RDE`` values, a boolean:
            * ``False``  no normalisation - default
            * ``True``  normalises w.r.t # possible vector permutations 
              (``c^m``  or ``(2c -1)^m-1`` if ``Fluct == True``).

        :rho:   - *If ``Typex == 'finesort'``, rho is the tuning parameter (default: 1)
 
    :See also:
        ``PermEn``, ``SyDyEn``, ``MSEn``.
    
    :References:
        [1] Mostafa Rostaghi and Hamed Azami,
            "Dispersion entropy: A measure for time-series analysis." 
            IEEE Signal Processing Letters 
            23.5 (2016): 610-614.
    
        [2] Hamed Azami and Javier Escudero,
            "Amplitude-and fluctuation-based dispersion entropy." 
            Entropy 
            20.3 (2018): 210.
    
        [3] Li Yuxing, Xiang Gao and Long Wang,
            "Reverse dispersion entropy: A new complexity measure for 
            sensor signal." 
            Sensors 
            19.23 (2019): 5203.
    
        [4] Wenlong Fu, et al.,
            "Fault diagnosis for rolling bearings based on fine-sorted 
            dispersion entropy and SVM optimized with mutation SCA-PSO."
            Entropy
            21.4 (2019): 404.
    
    """

    Sig = np.squeeze(Sig)
    N = Sig.shape[0]       
        
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(c,int) and (c > 1), "c:     must be an integer > 1"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Fluct,bool), "Fluct:    must be boolean - True or False"
    assert Typex.lower() in ['linear','kmeans','ncdf','finesort','equal'], \
            "Typex:    must be one of the following strings - \
            'linear', 'kmeans', 'ncdf', 'finesort', 'equal'"     
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"  
            
    if Typex.lower() == 'linear':
        Zi = np.digitize(Sig,np.arange(min(Sig),max(Sig),np.ptp(Sig)/c))
        
    elif Typex.lower() == 'kmeans':        
        Clux, Zx = kmeans2(Sig, c, iter=200)
        Zx += 1;  xx = np.argsort(Clux) + 1;     Zi = np.zeros(N);
        for k in range(1,c+1):
            Zi[Zx==xx[k-1]] = k;
        
    elif Typex.lower() == 'ncdf':  
        Zx = ndtr((Sig-np.mean(Sig))/np.std(Sig))
        Zi = np.digitize(Zx,np.arange(0,1,1/c))  
        
        del Zx
        
    elif Typex.lower() == 'finesort':        
        assert isinstance(rho,(float,int)) and rho>0 \
        and Typex.lower()=='finesort', "rho:  must be >0 and Typex must be 'finesort'"
        
        Zx = ndtr((Sig-np.mean(Sig))/np.std(Sig))
        Zi = np.digitize(Zx,np.arange(0,1,1/c))        
        Ym = np.zeros((N-(m-1)*tau, m))
        for n in range(m):
            Ym[:,n] = Zx[n*tau:N-((m-n-1)*tau)]
        
        Yi = np.floor(np.max(abs(np.diff(Ym)),axis=1)/(rho*np.std(abs(np.diff(Sig)))))         
        del Zx, Ym, rho
                
    elif Typex.lower() == 'equal':
        ix = np.argsort(Sig, kind='mergesort')
        xx = np.round(np.arange(0,2*N,N/c)).astype(int)
        Zi = np.zeros(N)
        for k in range(c):
            Zi[ix[xx[k]:xx[k+1]]] = k+1 
        del ix, xx, k        
            
    Zm = np.zeros((N-((m-1)*tau),m))   
    for n in np.arange(m):
        Zm[:,n] = Zi[n*tau:N-((m-n-1)*tau)]
    
    if Typex.lower() == 'finesort':
        Yi = np.expand_dims(Yi,axis=1)        
        Zm = np.hstack((Zm, Yi))        
        
    if Fluct:
        Zm = np.diff(Zm,axis=1)
        if m < 2:
           print('Warning:  Fluctuation-based Dispersion Entropy',
               'is undefined for m = 1.\n',
               'An embedding dimension (m) > 1 should be used.')       
           
    T  = np.unique(Zm,axis=0)
    Nx = T.shape[0]
    Counter = np.zeros(Nx)
       
    for n in range(Nx):
        Counter[n] = sum(np.any(Zm - T[n,:],axis=1)==False)

    Ppi = Counter[Counter!= 0]/Zm.shape[0]    
    if Fluct:
        RDE = sum((Ppi - (1/((2*c - 1)**(m-1))))**2);
    else:
        RDE = sum((Ppi - (1/(c**m)))**2);
        
    #RDE = sum(Ppi**2) - (1/Nx)
    Dispx = -sum(Ppi*np.log(Ppi)/np.log(Logx))    
    
    if round(sum(Ppi)) != 1:
        print('Potential error calculating probabilities')    
        
    if Norm:
        #Dispx = Dispx/(np.log(Nx)/np.log(Logx))
        #RDE = RDE/(1 - (1/(Nx)))                
        if Fluct:
            Dispx = Dispx/(np.log((2*c - 1)**(m-1))/np.log(Logx))
            RDE = RDE/(1 - (1/((2*c - 1)**(m-1))))
        else:
            Dispx = Dispx/(np.log(c**m)/np.log(Logx))
            RDE = RDE/(1 - (1/(c**m)))
                
    return Dispx, RDE
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Bidimensional Fuzzy Entropy function."""
import numpy as np    

def FuzzEn2D(Mat, m=None, tau=1, r=None, Fx='default', Logx=np.exp(1), Lock=True):
    """FuzzEn2D  estimates the bidimensional fuzzy entropy of a data matrix.
 
    .. code-block:: python
    
        Fuzz2D = FuzzEn2D(Mat) 
        
    Returns the bidimensional fuzzy entropy estimate (``Fuzz2D``) estimated for 
    the data matrix (``Mat``) using the default parameters: time delay = 1,
    fuzzy function (``Fx``) = ``'default'``, fuzzy function parameters (``r``) = (0.2*SD(Mat), 2),
    logarithm = natural, matrix template size = [floor(H/10) floor(W/10)] 
    (where H and W represent the height (rows) and width (columns) of the data matrix ``'Mat'``) 
    ** The minimum number of rows and columns of Mat must be > 10.
 
    .. code-block:: python
    
        Fuzz2D = FuzzEn2D(Mat, keyword = value, ...)

    Returns the bidimensional fuzzy entropy (``Fuzz2D``) estimates for the data
    matrix (``Mat``) using the specified 'keyword' arguments:
        :m:     - Template submatrix dimensions, an integer (for sub-matrix with same height and width) or a two-element vector of integers [height, width] with a minimum value > 1.  (default: [floor(H/10) floor(W/10)])
        :tau:   - Time Delay, a positive integer  (default: 1)
        :Fx:    - Fuzzy function name, one of the following strings:  
            {``'sigmoid'``, ``'modsampen'``, ``'default'``, ``'gudermannian'``, ``'bell'``, ``'triangular'``, ``'trapezoidal1'``, ``'trapezoidal2'``, ``'z_shaped'``, ``'gaussian'``, ``'constgaussian'``}
        :r:     - Fuzzy function parameters, a 1 element scalar or a 2 element vector of positive values. 
                The ``r`` parameters for each fuzzy  function are defined as follows:   [default: (0.2*SD(Mat), 2)]
                  - default:      [Tuple]
                      * r(1) = divisor of the exponential argument
                      * r(2) = argument exponent (pre-division)  
                  - sigmoid:      [Tuple]
                      * r(1) = divisor of the exponential argument
                      * r(2) = value subtracted from argument (pre-division)
                  - modsampen:    [Tuple]
                      * r(1) = divisor of the exponential argument
                      * r(2) = value subtracted from argument (pre-division)
                  - gudermannian:   
                      * r  = a scalar whose value is the numerator of  argument to gudermannian function: GD(x) = atan(tanh(r/x)). GD(x) is normalised to have a maximum value of 1.
                  - triangular:  
                      * r = a scalar whose value is the threshold (corner point) of the triangular function.
                  - trapezoidal1:  
                      * r = a scalar whose value corresponds to the upper (2r) and lower (r) corner points of the trapezoid.
                  - trapezoidal2:  [Tuple]
                      * r(1) = a value corresponding to the upper corner point of the trapezoid.
                      * r(2) = a value corresponding to the lower corner point of the trapezoid.
                  - z_shaped:  
                      * r = a scalar whose value corresponds to the upper (2r) and lower (r) corner points of the z-shape.
                  - bell:  
                      * r(1) = divisor of the distance value
                      * r(2) = exponent of generalized bell-shaped function
                  - gaussian:  
                      * r = a scalar whose value scales the slope of the Gaussian curve.
                  - constgaussian:  
                      * r = a scalar whose value defines the lower threshod and shape of the Gaussian curve.                    
                  - [DEPRICATED] linear:       
                      r  = an integer value. When r = 0, the argument of the exponential function is 
                      normalised between [0 1]. When r = 1, the minimuum value of the exponential argument is set to 0.    
                        
        :Logx:  - Logarithm base, a positive scalar    (default: natural)
        :Lock:  - By default, ``FuzzEn2D`` only permits matrices with a maximum  size of 128 x 128 to prevent RAM overload. 
                  e.g. For ``Mat`` = [200 x 200], ``m`` = 3, and ``tau`` = 1, ``FuzzEn2D`` 
                  creates a vector of 753049836 elements. To enable matrices greater than [128 x 128] elements, set ``Lock = False`` (default: True)
                  
                  ``CAUTION: unlocking the permitted matrix size may cause memory``
                  ``errors that could lead your Python IDE to crash.``
    
    :See also:
        ``SampEn2D``, ``DistEn2D``, ``FuzzEn``, ``XFuzzEn``, ``XMSEn``
    
    :References:
        [1] Luiz Fernando Segato Dos Santos, et al.,
            "Multidimensional and fuzzy sample entropy (SampEnMF) for
            quantifying H&E histological images of colorectal cancer."
            Computers in biology and medicine 
            103 (2018): 148-160.
    
        [2] Mirvana Hilal and Anne Humeau-Heurtier,
            "Bidimensional fuzzy entropy: Principle analysis and biomedical
            applications."
            41st Annual International Conference of the IEEE (EMBC) Society
            2019.
    
        [3] Hamed Azami, et al.
            "Fuzzy Entropy Metrics for the Analysis of Biomedical Signals: 
            Assessment and Comparison"
            IEEE Access
            7 (2019): 104833-104847
     """ 
    
    Mat = np.squeeze(Mat)
    NL,NW = Mat.shape     
    if m is None:
        m = np.array(Mat.shape)//10
    if r is None:
        r = (0.2*np.std(Mat), 2)

    assert Mat.ndim==2 and min(Mat.shape)>5 , \
    "Mat:   must be a 2D numpy array with height & width > 10"
    assert (isinstance(m,int) and m>1) or (isinstance(m, (np.ndarray,tuple))  
                                           and len(m)==2 and min(m)>1), \
    "m:     must be an integer > 1, or a 2 element tuple of integers > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) or ((r[0] >= 0) and len(r) ==2), "r:     must be 2 element tuple of positive values"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert Fx.lower() in ['default','sigmoid','modsampen','gudermannian', 'bell', 
                          'gaussian', 'constgaussian', 'triangular' , 
                          'trapezoidal1', 'trapezoidal2', 'z_shaped'] \
            and isinstance(Fx,str), """Fx:    must be one of the following strings -
            'default', 'sigmoid', 'modsampen', 'gudermannian', 'bell', 'z_shaped',
            'triangular', 'trapezoidal1','trapezoidal2','gaussian','constgaussian'"""
    assert isinstance(Lock,bool) and ((Lock==True and max(Mat.shape)<=128) or Lock==False), \
    "Lock:      To prevent memory storage errors, matrix width & length must \
    have <= 128 elements. To estimate FuzzEn2D for the current matrix (%d x %d) \
    change Lock to False. \
    CAUTION: unlocking the permitted matrix size may cause memory \
    errors that could lead your Python IDE to crash.."%(NW,NL)
    
    if isinstance(m,int):
        mL = int(m); mW = int(m)        
    else:
        mL = int(m[0]); mW = int(m[1])   
    
    if isinstance(r,tuple) and Fx.lower()=='linear':
        r = 0
        print('Multiple values for r entered. Default value (0) used.') 
    elif isinstance(r,tuple) and Fx.lower()=='gudermannian':
        r = r[0];
        print('Multiple values for r entered. First value used.')     
        
    Fun = globals()[Fx.lower()]
    NL = NL - mL*tau
    NW = NW - mW*tau
    X1 = np.zeros((NL*NW,mL,mW))
    X2 = np.zeros((NL*NW,mL+1,mW+1))
    p = 0
    for k in range(NL):        
        for n in range(NW):
            Temp2 = Mat[k:(mL+1)*tau+k:tau,n:(mW+1)*tau+n:tau]
            Temp1 = Temp2[:-1,:-1]
            X1[p,:,:] = Temp1 - np.mean(Temp1)
            X2[p,:,:] = Temp2 - np.mean(Temp2)
            p += 1            
    if p != NL*NW:
        print('Warning: Potential error with submatrix division.')        
    Ny = int(p*(p-1)/2)
    if Ny > 300000000:
        print('Warning: Number of pairwise distance calculations is ' + str(Ny))
    
    Y1 = np.zeros(p-1)
    Y2 = np.zeros(p-1)
    for k in range(p-1):
        Temp1 = Fun(np.max(abs(X1[k+1:,:,:] - X1[k,:,:]),axis=(1,2)),r)
        Y1[k] = np.sum(Temp1)        
        Temp2 = Fun(np.max(abs(X2[k+1:,:,:] - X2[k,:,:]),axis=(1,2)),r)
        Y2[k] = np.sum(Temp2) 
        
    Fuzz2D = -np.log(sum(Y2)/sum(Y1))/np.log(Logx)
    return Fuzz2D


def sigmoid(x,r):
    assert isinstance(r,tuple), 'When Fx = "Sigmoid", r must be a two-element tuple.'
    y = 1/(1 + np.exp((x-r[1])/r[0]))
    return y  
def default(x,r):   
    assert isinstance(r,tuple), 'When Fx = "Default", r must be a two-element tuple.'
    y = np.exp(-(x**r[1])/r[0])
    return y     
def modsampen(x,r):
    assert isinstance(r,tuple), 'When Fx = "Modsampen", r must be a two-element tuple.'
    y = 1/(1 + np.exp((x-r[1])/r[0]));
    return y    
def gudermannian(x,r):
    if r <= 0:
        raise Exception('When Fx = "Gudermannian", r must be a scalar > 0.')
    y = np.arctan(np.tanh(r/x))    
    y = y/np.max(y)    
    return y    
# def linear(x,r):    
#     if r == 0 and x.shape[0]>1:    
#         y = np.exp(-(x - min(x))/np.ptp(x))
#     elif r == 1:
#         y = np.exp(-(x - min(x)))
#     elif r == 0 and x.shape[0]==1:   
#         y = 0;
#     else:
#         print(r)
#         raise Exception('When Fx = "Linear", r must be 0 or 1')
#     return y
def triangular(x, r):
    assert np.size(r)==1, 'When Fx = "Triangular", r must be a scalar > 0.'
    y = 1 - (x/r)
    y[x > r] = 0
    return y
def trapezoidal1(x, r):
    assert np.size(r)==1, 'When Fx = "Trapezoidal1", r must be a scalar > 0.'
    y = np.zeros(np.shape(x))
    y[x <= r*2] = 2 - (x[x <= r*2]/r)
    y[x <= r] = 1
    return y
def trapezoidal2(x, r):
    assert isinstance(r,tuple), 'When Fx = "Trapezoidal2", r must be a two-element tuple.'
    y = np.zeros(np.shape(x))
    y[x <= max(r)] = 1 - (x[x <= max(r)] - min(r))/(max(r)-min(r))
    y[x <= min(r)] = 1
    return y
def z_shaped(x, r):
    assert np.size(r)==1, 'When Fx = "Z_shaped", r must be a scalar > 0.'
    y = np.zeros(np.shape(x))
    y[x <= 2*r] = 2*(((x[x <= 2*r] - 2*r)/r)**2)
    y[x <= 1.5*r] = 1 - (2*(((x[x <= 1.5*r] - r)/r)**2))
    y[x <= r] = 1
    return y
def bell(x, r):
    assert isinstance(r,tuple), 'When Fx = "Bell", r must be a two-element tuple.'
    y = 1/(1 + abs(x/r[0])**(2*r[1]))
    return y
def gaussian(x, r):
    assert np.size(r)==1, 'When Fx = "Gaussian", r must be a scalar > 0.'
    y = np.exp(-((x**2)/(2*(r**2))))
    return y
def constgaussian(x, r):
    assert np.size(r)==1, 'When Fx = "ConstGaussian", r must be a scalar > 0.'
    y = np.ones(np.shape(x))
    y[x > r] = np.exp(-np.log(2)*((x[x > r] - r)/r)**2)
    return y

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base composite Multiscale Entropy function."
import numpy as np 
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy

def cMSEn(Sig, Mbjx, Scales=3, RadNew=0, Refined=False, Plotx=False):
    """cMSEn  Returns the composite (or refined-composite) multiscale entropy of a univariate data sequence.

    .. code-block:: python 
        
        MSx, CI = cMSEn(Sig, Mobj) 
        
    Returns a vector of composite multiscale entropy values (``MSx``) for the data 
    sequence (``Sig``) using the parameters specified by the multiscale object 
    (``Mobj``) using the composite multiscale entropy method (cMSE) over 3 temporal
    scales. 
    
    .. code-block:: python 
        
        MSx, CI = cMSEn(Sig, Mobj, Refined = True) 
        
    Returns a vector of refined-composite multiscale entropy values (``MSx``) for the data 
    sequence (``Sig``) using the parameters specified by the multiscale object 
    (``Mobj``) using the refined-composite multiscale entropy method (rcMSE) over 3 temporal
    scales. When ``Refined == True``, the base entropy method must be ``SampEn`` or ``FuzzEn``.
    If the entropy method is ``SampEn``, cMSEn employs the method described in [5]. 
    If the entropy method is ``FuzzEn``, cMSEn employs the method described in [6].     
    
    .. code-block:: python
        
        MSx, CI = cMSEn(Sig, Mobj, keyword = value, ...)
        
    Returns a vector of composite multiscale entropy values (``MSx``) of the 
    data sequence (``Sig``) using the parameters specified by the multiscale 
    object (``Mobj``) and the following 'keyword' arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default: 3)
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the entropy specified by ``Mobj`` is ``SampEn`` or ``ApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Standard Deviation          - ``r*std(Ykj)``
                        * [2] Variance                    - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                     
        :Refined:  - Refined-composite MSEn method. When ``Refined == True`` and the entropy function specified by ``Mobj`` is ``SampEn`` or ``FuzzEn``, 
                     ``cMSEn`` returns the refined-composite multiscale entropy (rcMSEn) [default: False]
        :Plotx:    - When Plotx == True, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    :See also:
        ``MSobject``, ``MSEn``, ``rMSEn``, ``hMSEn``, ``XMSEn``, ``cXMSEn``, ``SampEn``, ``ApEn`` 
     
    :References:
        [1] Madalena Costa, Ary Goldberger, and C-K. Peng,
            "Multiscale entropy analysis of complex physiologic time series."
            Physical review letters
            89.6 (2002): 068102.
            
        [2] Vadim V. Nikulin, and Tom Brismar,
            "Comment on “Multiscale entropy analysis of complex physiologic
            time series”." 
            Physical review letters 
            92.8 (2004): 089803.
    
        [3] Madalena Costa, Ary L. Goldberger, and C-K. Peng. 
            "Costa, Goldberger, and Peng reply." 
            Physical Review Letters
            92.8 (2004): 089804.
    
        [4] Shuen-De Wu, et al.,
            "Time series analysis using composite multiscale entropy." 
            Entropy 
            15.3 (2013): 1069-1084.
    
        [5] Shuen-De Wu, et al.,
            "Analysis of complex time series using refined composite 
            multiscale entropy." 
            Physics Letters A 
            378.20 (2014): 1369-1374.
            
        [6] Hamed Azami et al.,
            "Refined multiscale fuzzy entropy based on standard deviation 
            for biomedical signal analysis"
            Med Biol Eng Comput 
            55 (2017):2037–2052

    
    """
  
    Mobj = deepcopy(Mbjx)    
    Sig = np.squeeze(Sig)    
    assert Sig.shape[0]>10 and Sig.ndim == 1, "Sig:   must be a numpy vector"    
    assert isinstance(Mobj,object), "Mobj:  must be a multiscale entropy \
    object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert isinstance(RadNew,int) and (np.isin(RadNew,np.arange(1,5)) \
                and Mobj.Func.__name__ in ['SampEn','ApEn']) or RadNew==0, \
    "RadNew:     must be 0, or an integer in range [1 4] with entropy function 'SampEn' or 'ApEn'"
    assert isinstance(Refined, bool) and ((Refined==True and Mobj.Func.__name__ in ['SampEn','FuzzEn'])
        or Refined==False), \
    "Refined:       must be a bool (True or False). If Refined==True, Mobj.Func must be SampEn or FuzzEn"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
    assert Mobj.Func.__name__[0].lower() != 'x', ("Base entropy estimator is a cross-entropy method." 
    "To perform composite multiscale CROSS-entropy estimation, use cXMSEn.")
    
    if Mobj.Func.__name__ in 'SampEn': Mobj.Kwargs['Vcp'] = False 
    
    MSx = np.zeros(Scales)    
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x: np.std(x)
        elif RadNew == 2:
            Rnew = lambda x: np.var(x)
        elif RadNew == 3:
            Rnew = lambda x: np.mean(abs(x-np.mean(x)))
        elif RadNew == 4:
            Rnew = lambda x: np.median(abs(x-np.median(x)))    
        
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Standard Deviation','Variance','Mean Abs Deviation',
                  'Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])            
            Cx = .2

    if Refined and Mobj.Func.__name__ == 'FuzzEn':
        Tx = 1
        if 'Logx' in Mobj.Kwargs.keys(): Logx = Mobj.Kwargs['Logx']
        else: Logx = np.e
        
    elif Refined and Mobj.Func.__name__ == 'SampEn':
        Tx = 0
        if 'Logx' in Mobj.Kwargs.keys(): Logx = Mobj.Kwargs['Logx']
        else: Logx = np.e

    else:  Tx = 0
    
    for T in range(1,Scales+1):
        Temp = modified(Sig,T,Tx)    
        
        N = T*(len(Temp)//T)
        Temp3 = np.zeros(T)
        Temp2 = np.zeros(T)                
        for k in range(T):        
            print(' .', end='')
            if RadNew > 0:
                Mobj.Kwargs.update({'r': Cx*Rnew(Temp[k::T])})       
                
            if Refined:
                _, Ma, Mb = Mobj.Func(Temp[k:N:T],**Mobj.Kwargs)
                Temp2[k] = Ma[-1]
                Temp3[k] = Mb[-1]
            else:
                Temp2 = Mobj.Func(Temp[k:N:T],**Mobj.Kwargs)
                if isinstance(Temp2,tuple):
                    if isinstance(Temp2[0],(int,float)):
                        Temp3[k] = Temp2[0]
                    else:
                        Temp3[k] = Temp2[0][-1]       
                elif isinstance(Temp2,(int,float)):
                    Temp3[k] = Temp2
                elif isinstance(Temp2,np.ndarray):
                    Temp3[k] = Temp2[-1]
                     
        if Refined and Tx==0:
            MSx[T-1] = -np.log(np.sum(Temp2)/np.sum(Temp3))/np.log(Logx)
        elif Refined and Tx==1:
            MSx[T-1] = -np.log(np.sum(Temp3)/np.sum(Temp2))/np.log(Logx)
        else:
            MSx[T-1] = np.mean(Temp3)
    
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')

    if Plotx:
        if Refined:
            strx = 'Refined-Composite'
        else:
            strx = 'Composite'
        
        figure()
        ax1 = axes()   
        ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
        ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
        ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_title('%s Multiscale %s'%(strx,Mobj.Func.__name__), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))       
        show()
    
    return MSx, CI

def modified(Z,sx, Tx):
    # Ns = len(Z) - sx +1
    # Y = np.zeros(Ns)
    # for k in range(Ns):
    #     Y[k] = np.mean(Z[k:k+sx])
    
    if Tx==1:
        Y = np.asarray([np.std(Z[x:x+sx]) for x in range(len(Z)-sx+1)])
    else: 
        Y = np.convolve(Z,np.ones(sx),'valid')/sx
        
    return Y




"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base corrected Cross Conditional Entropy function."""
import numpy as np 

def XCondEn(*Sig, m=2, tau=1, c=6, Logx=np.exp(1), Norm=False):
    """XCondEn  estimates the corrected cross-conditional entropy between two univariate data sequences.

    .. code-block:: python 
        
        XCond, SEw, SEz = XCondEn(Sig1, Sig2) 
        
    Returns the corrected cross-conditional entropy estimates (``XCond``) and the
    corresponding Shannon entropies (``m: SEw``, ``m+1: SEz``) for ``m`` = [1,2] 
    estimated for the data sequences contained in ``Sig1`` and ``Sig2`` using the default
    parameters:  embedding dimension = 2, time delay = 1, 
    number of symbols = 6, logarithm = natural
    ** Note: ``XCondEn`` is direction-dependent. Therefore, the order of the
    data sequences ``Sig1`` and ``Sig2`` matters. If ``Sig1`` is the
    sequence 'y', and ``Sig2`` is the sequence 'u', then ``XCond`` is
    the amount of information carried by y(i) when the pattern u(i) is found.
  
    .. code-block:: python

        XCond, SEw, SEz = XCondEn(Sig1, Sig2, keyword = value, ...)
        
    Returns the corrected cross-conditional entropy estimates (``XCond``) for
    the data sequences contained in ``Sig1`` and ``Sig2`` using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 1        [default: 2]
        :tau:   - Time Delay, a positive integer             [default: 1]
        :c:     - Number of symbols, an integer > 1          [default: 6]
        :Logx:  - Logarithm base, a positive scalar          [default: natural]
        :Norm:  - Normalisation of XCond value, one of the following integers:
            * [False]  no normalisation                  [default]
            * [True]   normalises w.r.t cross-Shannon entropy.  
 
    :See also:
        ``XFuzzEn``, ``XSampEn``, ``XApEn``, ``XPermEn``, ``CondEn``, ``XMSEn``
    
    :References:
        [1] Alberto Porta, et al.,
            "Conditional entropy approach for the evaluation of the 
            coupling strength." 
            Biological cybernetics 
            81.2 (1999): 119-129.
         
    """
  
    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])        
        assert len(S1)==len(S2),  "Sig1/Sig2:   Each sequence must be of equal length"

    N  = S1.shape[0]

    assert N>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(c,int) and (c > 1), "c:   must be an integer > 1"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"

    #S1 = (Sig[:,0]-np.mean(Sig[:,0]))/np.std(Sig[:,0])
    #S2 = (Sig[:,1]-np.mean(Sig[:,1]))/np.std(Sig[:,1])   
    
    S1 = (S1-np.mean(S1))/np.std(S1)
    S2 = (S2-np.mean(S2))/np.std(S2)
    
    Edges = np.arange(min(S1),max(S1),np.ptp(S1)/c)
    Sx1 = np.digitize(S1,Edges)   
    Edges = np.arange(min(S2),max(S2),np.ptp(S2)/c)
    Sx2 = np.digitize(S2,Edges)
       
    SEw = np.zeros(m-1)
    SEz = np.zeros(m-1)
    Prcm = np.zeros(m-1)
    Xi = np.zeros((N,m))
    
    for k in range(1,m):
        
        Nx = N-(k-1)*tau
        Xi[:Nx,-k] = Sx1[(k-1)*tau:]        
        Wi = np.inner(c**np.arange(k-1,-1,-1),Xi[:Nx,-k:])
        Zi = (c**k)*Sx2[(k-1)*tau:] + Wi   
        Pw, _ = np.histogram(Wi,np.arange(min(Wi)-.5,max(Wi)+1.5))
        Pz, _ = np.histogram(Zi,np.arange(min(Zi)-.5,max(Zi)+1.5))
        Prcm[k-1] = sum(Pz==1)/Nx
        
        if sum(Pw)!=Nx or sum(Pz)!=Nx:
            print('Warning: Potential error estimating probabilities.')
        Pw = Pw[Pw!=0]; Pw = Pw/N
        Pz = Pz[Pz!=0]; Pz = Pz/N
        SEw[k-1] = -sum(Pw*np.log(Pw)/np.log(Logx))
        SEz[k-1] = -sum(Pz*np.log(Pz)/np.log(Logx))
        del Pw, Pz, Wi, Zi
        
    Temp, _ = np.histogram(Sx2,c)
    Temp = Temp[Temp!=0]/N
    Sy = -np.sum(Temp*np.log(Temp)/np.log(Logx))
    XCond = SEz - SEw + Prcm*Sy
    XCond = np.hstack((Sy, XCond))
    if Norm:
        XCond = XCond/Sy
    
    return XCond, SEw, SEz

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


""" Base Multivariate Sample Entropy function."""
import numpy as np       

def MvSampEn(Data, m=None, tau=None, r=0.2, Norm=False, Logx=np.exp(1)):
    """MvSampEn  estimates the multivariate sample entropy of a multivariate dataset.

    .. code-block:: python
    
        MSamp, B0, Bt, B1 = MvSampEn(Data) 
    
    Returns the multivariate sample entropy estimate (``MSamp``) and the
    average number of matched delay vectors (``m``: ``B0``; joint total 
    ``m+1`` subspace: ``Bt``; all possible ``m+1`` subspaces: ``B1``),
    from the M multivariate sequences in ``Data`` using the default parameters: 
    embedding dimension = 2*np.ones(M), time delay = np.ones(M), radius threshold = 0.2,
    logarithm = natural, data normalization = false
    
    .. attention::
 
        The entropy value returned as ``MSamp`` is estimated using the "full" 
        method [i.e.  -log(Bt/B0)] which compares delay vectors across all possible ``m+1`` 
        expansions of the embedding space as applied in [1][2]. Contrary to
        conventional definitions of sample entropy, this method does not provide a
        lower bound of 0!!
        Thus, it is possible to obtain negative entropy values for multivariate 
        sample entropy, even for stochastic processes...
      
        Alternatively, one can calculate ``MSamp`` via the "naive" method, 
        which ensures a lower bound of 0, by using the average number of matched
        vectors for an individual ``m+1`` subspace (B1) [e.g. -log(B1(1)/B0)],
        or the average for all ``m+1`` subspaces [i.e. -log(mean(B1)/B0)].

    .. note::
      
        To maximize the number of points in the embedding process, this algorithm 
        uses N-max(m*tau) delay vectors and **not** N-max(m)*max(tau) as employed 
        in [1][2].
    

 
    .. code-block:: python
    
        MSamp, B0, Bt, B1 = MvSampEn(Data, keyword = value, ...)
        
    Returns the multivariate sample entropy estimates (``MSamp``) estimated
    from the M multivariate data sequences in ``Data`` using the specified 
    keyword  arguments:
        :Data:  - Multivariate dataset, NxM matrix of N (>10) observations (rows) and M (cols) univariate data sequences 
        :m:     - Embedding Dimension, a vector of M positive integers
        :tau:   - Time Delay, a vector of M positive integers
        :r:     - Radius Distance threshold, a positive scalar
        :Norm:  - Normalisation of all M sequences to unit variance, a boolean
        :Logx:  - Logarithm base, a positive scalar 

            
    :See also:
        ``SampEn``, ``XSampEn``, ``SampEn2D``, ``MSEn``, ``MvFuzzEn``, ``MvPermEn``.

    :References:
       [1] Ahmed Mosabber Uddin, Danilo P. Mandic
            "Multivariate multiscale entropy: A tool for complexity
            analysis of multichannel data."
            Physical Review E 84.6 (2011): 061918.

       [2] Ahmed Mosabber Uddin, Danilo P. Mandic
            "Multivariate multiscale entropy analysis."
            IEEE signal processing letters 19.2 (2011): 91-94.

    """
   
    Data = np.squeeze(Data)
    assert Data.shape[0]>10 and Data.ndim==2 and Data.shape[1]>1,  "Data:   must be an NxM numpy matrix where N>10 and M>1"
    N, Dn = Data.shape 
    if m is None:    m = 2*np.ones(Dn, dtype=int)
    if tau is None:  tau = np.ones(Dn, dtype=int)
    m = m.astype(int)
    tau = tau.astype(int) 
  
    # and np.issubdtype(m.dtype, np.integer)
    # and np.issubdtype(tau.dtype, np.integer)
    assert isinstance(m,np.ndarray) and all(m>0) and m.size==Dn and m.ndim==1, "m:     must be numpy vector of M positive integers"
    assert isinstance(tau,np.ndarray) and all(tau>0) and tau.size==Dn and tau.ndim==1, "tau:   must be numpy vector of M positive integers"
    assert isinstance(r,(int,float)) and (r>=0), "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be a Boolean"    

    if Norm: Data = Data/np.std(Data,axis=0)
    
    Nx = N - max((m-1)*tau)
    Ny = N - max(m*tau)       
    Vex = np.zeros((Nx,sum(m)))
    q = 0
    for k in range(Dn):
        for p in range(m[k]):
            Vex[:,q] = Data[p*tau[k]:Nx+p*tau[k],  k]
            q += 1
            
    Count0 = Distx(Vex,r)
    B0 = np.sum(Count0)/(Nx*(Nx-1)/2)
            
    B1 = np.zeros(Dn)
    Vez = np.inf*np.ones((1,sum(m)+1));
    Temp = np.cumsum(m)
    for k in range(Dn):
        Sig = np.expand_dims(Data[m[k]*tau[k]:Ny+m[k]*tau[k], k],1)
        Vey = np.hstack((Vex[:Ny, :Temp[k]], Sig, Vex[:Ny, Temp[k]:]))
        Vez = np.vstack((Vez, Vey))
        Count1 = Distx(Vey, r)
        B1[k] = np.sum(Count1)/(Ny*(Ny-1)/2)
    Vez = Vez[1:,:]
    Count1 = Distx(Vez, r)
    Bt = np.sum(Count1)/(Dn*Ny*((Dn*Ny)-1)/2)
       
    with np.errstate(divide='ignore', invalid='ignore'):
        Samp = -np.log(Bt/B0)/np.log(Logx) 
   
    return Samp, B0, Bt, B1    


def Distx(Vex, r):
    Nt = Vex.shape[0]
    Counter = np.zeros((Nt-1,Nt-1),dtype=bool)
    for x in range(Nt-1):
         Counter[x,x:] = np.all(abs(Vex[x+1:,:] - Vex[x,:]) <= r, axis=1)
     
    return Counter


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Bidimensional Distribution Entropy function."""
import numpy as np
from scipy.stats import skew

def DistEn2D(Mat, m=None, tau=1, Bins='Sturges', Logx=2, Norm=2, Lock=True):
    """DistEn2D  Estimates the bidimensional distribution entropy of a data matrix.
    
    .. code-block:: python
    
        Dist2D = DistEn2D(Mat) 
        
    Returns the bidimensional distribution entropy estimate (``Dist2D``)
    estimated for the data matrix (``Mat``) using the default parameters:
    time delay = 1, binning method = 'sturges', logarithm = natural, 
    matrix template size = [floor(H/10) floor(W/10)] (where H and W represent
    the height (rows) and width (columns) of the data matrix ``Mat``) 
    * The minimum dimension size of ``Mat`` must be > 10.
 
    .. code-block:: python
    
        Dist2D = DistEn2D(Mat, keyword = value, ...)
    
    Returns the bidimensional distribution entropy (``Dist2D``) estimate for 
    the data matrix (Mat) using the specified 'keyword' arguments:
        :m:     - Template submatrix dimensions, an integer scalar (for sub-matrix with same height and width) or a two-element tuple of integers [height, width] with a minimum value > 1. (default: [floor(H/10) floor(W/10)])
        :tau:   - Time Delay, a positive integer  (default: 1)
        :Bins:  - Histogram bin selection method for distance distribution, one of the following:
            * an integer > 1 indicating the number of bins, 
            * or one of the following strings {``'sturges'``, ``'sqrt'``, ``'rice'``, ``'doanes'``} [default: 'sturges']
              
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 
        :Norm:  - Normalisation of Dist2D value, one of the following integers:
            * [0]  no normalisation.
            * [1]  normalises values of data matrix (Mat) to range [0 1].
            * [2]  normalises values of data matrix (Mat) to range [0 1], and normalises the distribution entropy value (``Dist2D``) w.r.t the number of histogram bins.  [default]
            * [3]  normalises the distribution entropy value (``Dist2D``) w.r.t the number of histogram bins, without normalising  data matrix values.
        :Lock:  - By default, ``DistEn2D`` only permits matrices with a maximum size of 128 x 128 to prevent RAM overload. 
              e.g. For ``Mat`` = [200 x 200], ``m`` = 3, and ``tau`` = 1, ``DistEn2D`` 
              creates a vector of 753049836 elements. To enable matrices greater than [128 x 128] elements, set ``Lock = False`` (default: True)
              
              ``CAUTION: unlocking the permitted matrix size may cause memory``
              ``errors that could lead your Python IDE to crash.``
    
    :See also:
        ``DistEn``, ``XDistEn``, ``SampEn2D``, ``FuzzEn2D``, ``MSEn``
    
    :References:
        [1] Hamed Azami, Javier Escudero and Anne Humeau-Heurtier,
            "Bidimensional distribution entropy to analyze the irregularity
            of small-sized textures."
            IEEE Signal Processing Letters 
            24.9 (2017): 1338-1342.
    
    """
        
    Mat = np.squeeze(Mat)
    NL,NW = Mat.shape    
    if m is None:
        m = np.array(Mat.shape)//10
    if Logx == 0:
        Logx = np.exp(1)
           
    assert Mat.ndim==2 and min(Mat.shape)>10 , \
    "Mat:   must be a 2D numpy array with height & width > 10"
    assert (isinstance(m,int) and m>1) \
        or (isinstance(m, (np.ndarray,tuple)) and len(m)==2 and min(m)>1), \
    "m:     must be an integer > 1, or a 2 element tuple of integers > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"     
    assert (isinstance(Bins,int) and Bins > 1) \
        or (Bins.lower() in ['sturges','sqrt','rice','doanes']), \
        "Bins:    an integer > 1, or can be one of the following strings - \
        'sturges', 'sqrt', 'rice', 'doanes'"
    assert isinstance(Norm,int) and np.isin(Norm,np.arange(4)), "Norm:     must be integer in range [0 4]"   
    assert isinstance(Lock,bool) and ((Lock==True and max(Mat.shape)<=128) or Lock==False), \
    "Lock:      To prevent memory storage errors, matrix width & length must \
    have <= 128 elements. To estimate DistEn2D for the current matrix (%d x %d) \
    change Lock to False. \
    CAUTION: unlocking the permitted matrix size may cause memory \
    errors that could lead your Python IDE to crash.."%(NW,NL)
    
    if isinstance(m,int):
        mL = int(m); mW = int(m)        
    else:
        mL = int(m[0]); mW = int(m[1])   
           
    if Norm == 1 or Norm == 2:
        Mat = (Mat - np.min(Mat))/np.ptp(Mat[:])
   
    NL = NL - (mL - 1)*tau
    NW = NW - (mW - 1)*tau
    X = np.zeros((NL*NW,mL,mW))
    p = 0
    for k in range(0,NL):        
        for n in range(0,NW):
              X[p,:,:] = Mat[k:mL*tau+k:tau,n:mW*tau+n:tau]
              p += 1              
    if p != NL*NW:
        print('Warning: Potential error with submatrix division.')        
    Ny = int(p*(p-1)/2)
    if Ny > 300000000:
        print('Warning: Number of pairwise distance calculations is ' + str(Ny))
    
    Y = np.zeros(Ny)
    for k in range(1,p):
        Ix = (int((k-1)*(p - k/2)), int(k*(p-((k+1)/2))))        
        Y[Ix[0]:Ix[1]] = np.max(abs(X[k:,:,:] - X[k-1,:,:]),axis=(1,2))
     
    if isinstance(Bins, str):
        if Bins.lower() == 'sturges':
            Bx = np.ceil(np.log2(Ny) + 1)
        elif Bins.lower() == 'rice':
            Bx = np.ceil(2*(Ny**(1/3)))
        elif Bins.lower() == 'sqrt':
            Bx = np.ceil(np.sqrt(Ny))
        elif Bins.lower()== 'doanes':
            sigma = np.sqrt(6*(Ny-2)/((Ny+1)*(Ny+3)))
            Bx = np.ceil(1+np.log2(Ny)+np.log2(1+abs(skew(Y)/sigma)))
        else:
            raise Exception('Please enter a valid binning method')               
    else:
        Bx = Bins
        
    By = np.linspace(min(Y),max(Y),int(Bx+1))
    Ppi,_ = np.histogram(Y,By)        
    Ppi = Ppi/Ny
    if round(sum(Ppi),6) != 1:
        print('Warning: Potential error estimating probabilities (p = ' +str(np.sum(Ppi))+ '.')
        Ppi = Ppi[Ppi!=0]
    elif any(Ppi==0):
        print('Note: '+str(sum(Ppi==0))+'/'+str(len(Ppi))+' bins were empty')
        Ppi = Ppi[Ppi!=0]
           
    Dist2D = -sum(Ppi*np.log(Ppi)/np.log(Logx))
    if Norm >= 2:
        Dist2D = Dist2D/(np.log(Bx)/np.log(Logx))
      
    return Dist2D

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


""" Base Multivariate Permutation Entropy function."""
import numpy as np       
# from itertools import permutations as prm 
from scipy.signal import hilbert

def MvPermEn(Data, m=None, tau=None, Typex=None, tpx=-1, Norm=False, Logx=2):
    """MvPermEn  estimates the multivariate permutation entropy of a multivariate dataset.

    .. code-block:: python
    
        MPerm, MPnorm = MvPermEn(Data) 
    
    Returns the multivariate permutation entropy estimate (``MPerm``) and
    the normalized permutation entropy for the M multivariate sequences in
    ``Data`` using the default parameters:
    embedding dimension = 2*np.ones(M,1), time delay = np.ones(M,1), 
    logarithm = 2, normalisation = w.r.t #symbols (sum(``m-1``))
    
    .. attention::

        The multivariate permutation entropy algorithm implemented here uses
        multivariate embedding based on Takens' embedding theorem, and follows
        the methods for multivariate entropy estimation through shared spatial 
        reconstruction as originally presented by Ahmed & Mandic [1]. 
           
        This function does **NOT** use the multivariate permutation entropy 
        algorithm of Morabito et al. (Entropy, 2012) where the entropy values of 
        individual univariate sequences are averaged because such methods do not
        follow the definition of multivariate embedding and therefore do not
        consider cross-channel statistical complexity.
        
    .. note::
           
        To maximize the number of points in the embedding process, this
        algorithm uses N-max(tau*m) delay vectors and **not** N-max(m)*max(tau)
        as employed in [1].

 
    .. code-block:: python
    
        MPerm, MPnorm = MvPermEn(Data, keyword = value, ...)
        
    Returns the multivariate permutation entropy estimate (``MPerm``) for
    the M multivariate data sequences in ``Data`` using the specified keyword arguments:
        :Data:  - Multivariate dataset, NxM matrix of N (>10) observations (rows) and M (cols) univariate data sequences 
        :m:     - Embedding Dimension, a vector of M positive integers
        :tau:   - Time Delay, a vector of M positive integers
        :Typex: - Permutation entropy variation, can be one of the following strings:
               {``'modified'``, ``'ampaware'``, ``'weighted'``, ``'edge'``, ``'phase'``}
                See the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_ for more info on MvPermEn variants.    
        :tpx:   - Tuning parameter for associated permutation entropy variation.
             :ampaware:  ``tpx`` is the A parameter, a value in range [0 1] (default = 0.5)
             :edge:      ``tpx`` is the r sensitivity parameter, a scalar > 0 (default = 1)
             :phase:     ``tpx`` is the option to unwrap the phase angle of Hilbert-transformed signal, either [] or 1 (default = 0)
        :Norm:  - Normalisation of MPnorm value, a boolean operator:
             :false:  normalises w.r.t log(# of permutation symbols [sum(m)-1]) - default
             :true:   normalises w.r.t log(# of all possible permutations [sum(m)!])
        :Logx:  - Logarithm base, a positive scalar (defualt = 2; enter 0 for  natural logarithm)

    
    :See also:
        ``PermEn``, ``PermEn2D``, ``XPermEn``, ``MSEn``, ``MvFuzzEn``, ``MvSampEn``.
    
    :References:
      [1] Ahmed Mosabber Uddin, Danilo P. Mandic
            "Multivariate multiscale entropy: A tool for complexity
            analysis of multichannel data."
            Physical Review E 84.6 (2011): 061918.
    
      [2] Christoph Bandt and Bernd Pompe, 
            "Permutation entropy: A natural complexity measure for time series." 
            Physical Review Letters,
            88.17 (2002): 174102.
    
      [3] Chunhua Bian, et al.,
            "Modified permutation-entropy analysis of heartbeat dynamics."
            Physical Review E
            85.2 (2012) : 021906
    
      [4] Bilal Fadlallah, et al.,
            "Weighted-permutation entropy: A complexity measure for time 
            series incorporating amplitude information." 
            Physical Review E 
            87.2 (2013): 022911.
    
      [5] Hamed Azami and Javier Escudero,
            "Amplitude-aware permutation entropy: Illustration in spike 
            detection and signal segmentation." 
            Computer methods and programs in biomedicine,
            128 (2016): 40-51.
    
      [6] Zhiqiang Huo, et al.,
            "Edge Permutation Entropy: An Improved Entropy Measure for 
            Time-Series Analysis," 
            45th Annual Conference of the IEEE Industrial Electronics Soc,
            (2019), 5998-6003
    
      [7] Maik Riedl, Andreas Müller, and Niels Wessel,
            "Practical considerations of permutation entropy." 
            The European Physical Journal Special Topics 
            222.2 (2013): 249-262.
    
      [8] Kang Huan, Xiaofeng Zhang, and Guangbin Zhang,
           "Phase permutation entropy: A complexity measure for nonlinear time
           series incorporating phase information."
           Physica A: Statistical Mechanics and its Applications
           568 (2021): 125686.
      """
   
    Data = np.squeeze(Data)
    assert Data.shape[0]>10 and Data.ndim==2 and Data.shape[1]>1,  "Data:   must be an NxM numpy matrix where N>10 and M>1"
    N, Dn = Data.shape 
    if m is None:    m = 2*np.ones(Dn, dtype=int)
    if tau is None:  tau = np.ones(Dn, dtype=int)
    m = m.astype(int)
    tau = tau.astype(int) 
  
    assert isinstance(m,np.ndarray) and all(m>0) and m.size==Dn and m.ndim==1, "m:     must be numpy vector of M positive integers"
    assert isinstance(tau,np.ndarray) and all(tau>0) and tau.size==Dn and tau.ndim==1, "tau:   must be numpy vector of M positive integers"
    assert (isinstance(Typex,str) and Typex.lower() in ['modified','ampaware','weighted','edge','phase']) \
           or Typex is None, "Typex:    must be one of the following strings - 'modified','ampaware','weighted','edge', 'phase'"
    assert isinstance(tpx,(int,float)), \
          "tpx:   the value of tpx relates to 'Typex'. Type help(EntropyHub.MvPermEn) for further info on the 'tpx' value."   
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - False or True"
  
    if Typex is None: Typex = ''
    if Typex.lower() == 'phase':
        Data = np.angle(hilbert(Data,axis=0))
        if tpx == 1:
            Data = np.unwrap(Data)      
    if Logx == 0: Logx = np.exp(1)
    
    Nx = N - max((m-1)*tau)
    Sx = np.zeros((Nx,sum(m)))               
    q = 0 
    for k in range(Dn):
        for p in range(m[k]):
            Sx[:,q] = Data[p*tau[k]:Nx+p*tau[k],  k]
            q += 1
        
    Temp = np.argsort(Sx,axis=1)
    # WAY TOO SLOW - Px = np.asarray(list(prm(np.arange(sum(m)))))
    Px = np.unique(Temp,axis=0)
    Counter = np.zeros(Px.shape[0])
                    
    if Typex.lower() == 'modified':
        Tx = np.diff(np.sort(Sx,axis=1),axis=1)==0    
        for km in range(sum(m)-1):
            Temp[Tx[:,km],km+1] = Temp[Tx[:,km],km]                    

        if Tx.any()>0:   Px = np.vstack((Px,np.unique(Temp[np.any(Tx,axis=1),:],axis=0)))
            
        Counter = np.zeros(Px.shape[0])
        for n in range(Px.shape[0]):
            Counter[n] = np.sum(np.any(Temp - Px[n,:],axis=1)==False)                        
        Counter = Counter[Counter!= 0]
        Ppi = Counter/np.sum(Counter)
        del Tx

    elif Typex.lower() == 'weighted':
        Wj = np.var(Sx,axis=1)                
        for n in range(Px.shape[0]):
            Counter[n] = np.sum(Wj[np.any(Temp - Px[n,:],axis=1)==False])            
        Counter = Counter[Counter!= 0]
        Ppi = Counter/np.sum(Wj)
        del Wj, n

    elif Typex.lower() == 'ampaware':
        if tpx ==-1:
            tpx = 0.5
        elif tpx < 0 or tpx > 1:
            raise Exception('The A parameter (tpx) must be in the range [0 1]')
                              
        AA = np.sum(abs(Sx),axis=1)    
        AB = np.sum(abs(np.diff(Sx,axis=1)),axis=1)               
        Ax = ((tpx*AA/sum(m)) + (1-tpx)*AB/(sum(m)-1))                
        for n in range(Px.shape[0]):
            Counter[n] = np.sum(Ax[np.any(Temp - Px[n,:],axis=1)==False]) # Does Px need 2nd dim?
        Counter = Counter[Counter!= 0]
        Ppi = Counter/np.sum(Ax)                
        del AA, AB, Ax
  
    elif Typex.lower() == 'edge':
        if tpx == -1:
            tpx = 1
        elif tpx <= 0:
            raise Exception('r sensitivity parameter (tpx) must be > 0')
        
        for n in range(Px.shape[0]):
            Tx = np.diff(Sx[np.any(Temp - Px[n],axis=1)==False,:],axis=1)
            Counter[n] = np.sum(np.mean(np.hypot(Tx,1),axis=1)**tpx)
        Counter = Counter[Counter!= 0]
        Ppi = Counter/np.sum(Counter)               
        
    else:
        for n in range(Px.shape[0]):
            Counter[n] = sum(np.any(Temp - Px[n,:],axis=1)==False)            
        Counter = Counter[Counter!=0]
        Ppi = Counter/sum(Counter)
    
    if np.round(np.sum(Ppi),5) != 1:
        print('Warning: Potential error with probability calculation')
           
    with np.errstate(divide='ignore', invalid='ignore'):
        MPerm = -np.sum(Ppi*(np.log(Ppi)/np.log(Logx)))
    
    if Norm:
        MPnorm = MPerm/(np.log(np.math.factorial(sum(m)))/np.log(Logx))
    else:   MPnorm = MPerm/(sum(m)-1)
    
    return MPerm, MPnorm


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base heirarchical Multiscale Entropy function."
import numpy as np   
from matplotlib.pyplot import figure, show, subplot
import matplotlib.gridspec as gridspec
from copy import deepcopy

def hMSEn(Sig, Mbjx, Scales=3, RadNew=0, Plotx=False):
    """hMSEn  returns the hierarchical entropy of a univariate data sequence.
 
    .. code-block:: python
    
        MSx, Sn, CI = hMSEn(Sig, Mobj) 
        
    Returns a vector of entropy values (``MSx``) calculated at each node in the
    hierarchical tree, the average entropy value across all nodes at each 
    scale (``Sn``), and the complexity index (``CI``) of the hierarchical tree
    (i.e. ``sum(Sn)``) for the data sequence (``Sig``) using the parameters specified
    by the multiscale object (Mobj) over 3 temporal scales (default).
    The entropy values in MSx are ordered from the root node (S_00) to the
    Nth subnode at scale T (S_TN): i.e. S_00, S_10, S_11, S_20, S_21, S_22,
    S_23, S_30, S_31, S_32, S_33, S_34, S_35, S_36, S_37, S_40, ... , S_TN.
    The average entropy values in Sn are ordered in the same way, with the
    value of the root node given first: i.e. S0, S1, S2, ..., ST
     
    .. code-block:: python
    
        MSx, Sn, CI = hMSEn(Sig, Mobj, keyword = value, ...)
    
    Returns a vector of entropy values (``MSx``) calculated at each node in the
    hierarchical tree, the average entropy value across all nodes at each 
    scale (``Sn``), and the complexity index (``CI``) of the entire hierarchical
    tree for the data sequence (Sig) using the following 'keyword' arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default = 3)  At each scale (T), entropy is estimated for 2^(T-1) nodes.
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the entropy specified by ``Mobj`` is ``SampEn`` or ``ApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Standard Deviation          - ``r*std(Ykj)``
                        * [2] Variance                    - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                     
        :Plotx:    - When ``Plotx == True``, returns a plot of the average entropy value at each time scale (i.e. the multiscale entropy curve) and a network graph showing the entropy value of each node in the hierarchical tree decomposition.  (default: False)
    
    :See also:
        ``MSobject``, ``MSEn``, ``rMSEn``, ``cMSEn``, ``XMSEn``, ``hXMSEn``, ``rXMSEn``, ``cXMSEn``
    
    :References:
        [1] Ying Jiang, C-K. Peng and Yuesheng Xu,
            "Hierarchical entropy analysis for biological signals."
            Journal of Computational and Applied Mathematics
            236.5 (2011): 728-742.
    
    """
    
    Mobj = deepcopy(Mbjx)    
    Sig = np.squeeze(Sig)    
    assert Sig.shape[0]>10 and Sig.ndim == 1, "Sig:   must be a numpy vector"    
    assert isinstance(Mobj,object), "Mobj:  must be a multiscale entropy \
    object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert isinstance(RadNew,int) and (np.isin(RadNew,np.arange(1,5)) and Mobj.Func.__name__ in \
    ['SampEn','ApEn']) or RadNew==0, "RadNew:     must be an integer in range \
    [1 4] and entropy function must be 'SampEn' or 'ApEn'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
    assert Mobj.Func.__name__[0].lower() != 'x', ("Base entropy estimator is a cross-entropy method." 
    "To perform heirarchical multiscale CROSS-entropy estimation, use hXMSEn.")
    
    if Mobj.Func.__name__ in 'SampEn': Mobj.Kwargs['Vcp'] = False
    
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x: np.std(x)
        elif RadNew == 2:
            Rnew = lambda x: np.var(x)
        elif RadNew == 3:
            Rnew = lambda x: np.mean(abs(x-np.mean(x)))
        elif RadNew == 4:
            Rnew = lambda x: np.median(abs(x-np.median(x)))    
        
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Standard Deviation','Variance','Mean Abs Deviation',
                  'Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])            
            Cx = .2
    
    XX,N = Hierarchy(Sig,Scales)
    MSx = np.zeros(XX.shape[0])
    for T in range(XX.shape[0]):
        print(' .', end='')
        Temp = XX[T,:int(N/(2**(int(np.log2(T+1)))))]        
        if RadNew:
            Mobj.Kwargs.update({'r': Cx*Rnew(Temp)})            
        Temp2 = Mobj.Func(Temp,**Mobj.Kwargs)       
        
        if isinstance(Temp2,tuple):
            if isinstance(Temp2[0],(int,float)):
                MSx[T] = Temp2[0]
            else:
                MSx[T] =Temp2[0][-1]
        elif isinstance(Temp2,(int,float)):
            MSx[T] = Temp2
        elif isinstance(Temp2,np.ndarray):
            MSx[T] = Temp2[-1]
                                        
    Sn = np.zeros(Scales)
    for t in range(Scales):
        Sn[t] = np.mean(MSx[(2**t)-1:(2**(t+1))-1])
       
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')
    print('\n')
    
    if Plotx:
        
        figure()
        G = gridspec.GridSpec(10, 1)    
        ax1 = subplot(G[:2, :])
        ax1.plot(np.arange(1,Scales+1), Sn, color=(8/255, 63/255, 77/255), linewidth=3)
        ax1.scatter(np.arange(1,Scales+1), Sn, 60, color=(1,0,1))
        ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_title('Hierarchical Multiscale %s Entropy'%(Mobj.Func.__name__), 
                      fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))         

        N = 2**(Scales-1)
        x = np.zeros(2*N - 1,dtype=int)  
        x[0] = N;        
        y = Scales*(Scales - np.log2(np.arange(1,2*N))//1)
        for k in range(1,2*N):
            Q = int(np.log2(k)//1)
            P = int((k)//2)-1            
            if k>1:              
                if k%2:
                    x[k-1] =  x[P] + N/(2**Q)
                else:
                    x[k-1] =  x[P] - N/(2**Q)
                    
        Edges = np.vstack((np.repeat(np.arange(1,N),2),np.arange(2,2*N))).transpose() - 1
        labx = ["".join(k) for k in np.round(MSx,3).astype(str)]
        ax2 = subplot(G[3:,:])   
        ax2.scatter(x,y,s=100*(MSx-min(MSx)+1)/(abs(min(MSx))+1),color=(1,0,1))
        for k in range(len(x)-1):            
            ax2.plot(x[Edges[k,:]],y[Edges[k,:]],color=(8/255,63/255,77/255),linewidth=2.5)
            ax2.annotate(labx[k],(x[k],y[k]))  
        ax2.annotate(labx[-1],(x[-1],y[-1]))
        show()
    
    return MSx, Sn, CI


def Hierarchy(Z,sx):
    N = int(2**np.floor(np.log2(len(Z))))
    if np.log2(len(Z))%1 != 0:
        print('Only first %d samples were used in hierarchical decomposition. \
            \nThe last %d samples of the data sequence were ignored.'%(N,len(Z)-N))
    if N/(2**(sx-1)) < 8:
       raise Exception('Data length (%d) is too short to estimate entropy at the lowest' \
                      ' subtree. Consider reducing the number of scales.'%N) 
    
    U = np.zeros(((2**sx)-1,N))
    U[0,:] = Z[:N]
    p=1
    for k in range(sx-1):
        for n in range(2**k):
            Temp = U[(2**k)+n-1,:]        
            # U[p,1:N//2]  = (Temp[:-2:2] + Temp[1:-1:2])/2
            # U[p+1,1:N//2]= (Temp[:-2:2] - Temp[1:-1:2])/2
            U[p,:N//2]  = (Temp[::2] + Temp[1::2])/2
            U[p+1,:N//2]= (Temp[::2] - Temp[1::2])/2
            p+=2
    return U,N

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base Multiscale Entropy function."
import numpy as np   
from PyEMD import EMD
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy
# from scipy.stats import moment

def MSEn(Sig, Mbjx, Scales=3, Methodx='coarse', RadNew=0, Plotx=False):
    """MSEn  Returns the multiscale entropy of a univariate data sequence.
    
    .. code-block:: python
    
        MSx,CI = MSEn(Sig, Mobj) 
    
    Returns a vector of multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequence ``Sig`` using the parameters specified 
    by the multiscale object (``Mobj``) over 3 temporal scales with coarse-
    graining (default). 
  
    .. code-block:: python
    
        MSx,CI = MSEn(Sig, Mobj, keyword = value, ...)
        
    Returns a vector of multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequence ``Sig`` using the parameters specified by
    the multiscale object (``Mobj``) and the following ``keyword`` arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default: 3)
        :Methodx:  - Graining method, one of the following: [default: ``'coarse'``]  {``'coarse'``, ``'modified'``, ``'imf'`` , ``'timeshift'`` , ``'generalized'``}
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the entropy specified by ``Mobj`` is ``SampEn`` or ``ApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Standard Deviation          - ``r*std(Ykj)``
                        * [2] Variance                    - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                         
        :Plotx:    - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    For further info on these graining procedures see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
    
    :See also:
        ``MSobject``, ``rMSEn``, ``cMSEn``, ``hMSEn``, ``XMSEn``, ``cXMSEn``, ``hXMSEn``, ``SampEn``
    
    :References:
        [1] Madalena Costa, Ary Goldberger, and C-K. Peng,
            "Multiscale entropy analysis of complex physiologic time series."
            Physical review letters
            89.6 (2002): 068102.
    
        [2] Vadim V. Nikulin, and Tom Brismar,
            "Comment on “Multiscale entropy analysis of complex physiologic
            time series”." 
            Physical review letters 
            92.8 (2004): 089803.
    
        [3] Madalena Costa, Ary L. Goldberger, and C-K. Peng. 
            "Costa, Goldberger, and Peng reply." 
            Physical Review Letters
            92.8 (2004): 089804.
    
        [4] Madalena Costa, Ary L. Goldberger and C-K. Peng,
            "Multiscale entropy analysis of biological signals." 
            Physical review E 
            71.2 (2005): 021906.
    
        [5] Ranjit A. Thuraisingham and Georg A. Gottwald,
            "On multiscale entropy analysis for physiological data."
            Physica A: Statistical Mechanics and its Applications
            366 (2006): 323-332.
    
        [6] Meng Hu and Hualou Liang,
            "Intrinsic mode entropy based on multivariate empirical mode
            decomposition and its application to neural data analysis." 
            Cognitive neurodynamics
            5.3 (2011): 277-284.
    
        [7] Anne Humeau-Heurtier 
            "The multiscale entropy algorithm and its variants: A review." 
            Entropy 
            17.5 (2015): 3110-3123.
    
        [8] Jianbo Gao, et al.,
            "Multiscale entropy analysis of biological signals: a 
            fundamental bi-scaling law." 
            Frontiers in computational neuroscience 
            9 (2015): 64.
    
        [9] Paolo Castiglioni, et al.,
            "Multiscale Sample Entropy of cardiovascular signals: Does the
            choice between fixed-or varying-tolerance among scales 
            influence its evaluation and interpretation?." 
            Entropy
            19.11 (2017): 590.
    
        [10] Tuan D Pham,
            "Time-shift multiscale entropy analysis of physiological 
            signals." 
            Entropy 
            19.6 (2017): 257.
    
        [11] Hamed Azami and Javier Escudero,
            "Coarse-graining approaches in univariate multiscale sample 
            and dispersion entropy." 
            Entropy 20.2 (2018): 138.
            
        [12] Madalena Costa and Ary L. Goldberger,
            "Generalized multiscale entropy analysis: Application to quantifying 
            the complex volatility of human heartbeat time series." 
            Entropy 17.3 (2015): 1197-1203.
    
    """
    
    Mobj = deepcopy(Mbjx)
    
    Sig = np.squeeze(Sig)    
    Chk = ['coarse','modified','imf','timeshift','generalized']    
    assert Sig.shape[0]>10 and Sig.ndim == 1, "Sig:   must be a numpy vector"    
    assert isinstance(Mobj,object), "Mobj:  must be a multiscale entropy \
    object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert Methodx.lower() in Chk, "Methodx:  must be one of the following string names- \
    'coarse', 'modified' , 'imf', 'timeshift', 'generalized'"    
    assert isinstance(RadNew,int) and (np.isin(RadNew,np.arange(1,5)) \
                and Mobj.Func.__name__ in ['SampEn','ApEn']) or RadNew==0, \
    "RadNew:     must be 0, or an integer in range [1 4] with entropy function 'SampEn' or 'ApEn'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
    assert Mobj.Func.__name__[0].lower() != 'x', ("Base entropy estimator is a cross-entropy method." 
    "To perform multiscale CROSS-entropy estimation, use XMSEn.")
    
    if Mobj.Func.__name__ in 'SampEn': Mobj.Kwargs['Vcp'] = False 
    
    if Methodx.lower()=='imf':
        #EMD(total_power_thr=20)
        EMD().FIXE = 100
        EMD().FIXE_H = 100
        Sig = EMD().emd(Sig,max_imf=Scales-1)    
        assert Sig.shape[0]==Scales

    Func2 = globals()[Methodx.lower()]
    MSx = np.zeros(Scales)
    
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x: np.std(x)
        elif RadNew == 2:
            Rnew = lambda x: np.var(x)
        elif RadNew == 3:
            Rnew = lambda x: np.mean(abs(x-np.mean(x)))
        elif RadNew == 4:
            Rnew = lambda x: np.median(abs(x-np.median(x)))    
        
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Standard Deviation','Variance','Mean Abs Deviation',
                  'Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])            
            Cx = .2

    for T in range(1,Scales+1):
        print(' .', end='')
        Temp = Func2(Sig,T)        
        if Methodx.lower() == 'timeshift':
            Tempx = np.zeros(T)
            for k in range(T):
                if RadNew > 0:
                    Mobj.Kwargs.update({'r': Cx*Rnew(Temp[k,:])})                                    
                Tempy = Mobj.Func(Temp[k,:],**Mobj.Kwargs)
                
                if isinstance(Tempy,tuple):
                    if isinstance(Tempy[0],(int,float)):
                        Tempx[k] = Tempy[0]
                    else:
                        Tempx[k] =Tempy[0][-1]
                elif isinstance(Tempy,(int,float)):
                    Tempx[k] = Tempy
                elif isinstance(Tempy,np.ndarray):
                    Tempx[k] = Tempy[-1]
            
            Temp2 = np.mean(Tempx)
            
        else:
            if RadNew > 0:
                Mobj.Kwargs.update({'r': Cx*Rnew(Temp)})            
            Tempx = Mobj.Func(Temp,**Mobj.Kwargs)
                        
            if isinstance(Tempx,tuple):
                if isinstance(Tempx[0],(int,float)):
                    Temp2 = Tempx[0]
                else:
                    Temp2 =Tempx[0][-1]
            elif isinstance(Tempx,(int,float)):
                Temp2 = Tempx
            elif isinstance(Tempx,np.ndarray):
                Temp2 = Tempx[-1]
                    
        MSx[T-1] = Temp2
            
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')

    if Plotx:
       figure()
       ax1 = axes()   
       ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
       ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
       ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_title('Multiscale %s (%s-graining method)'%(Mobj.Func.__name__,Methodx), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))      
       show()
    
    return MSx, CI


def coarse(Z,sx):
    Ns = len(Z)//sx
    Y = np.mean(np.reshape(Z[:sx*Ns],(Ns,sx)),axis=1)
    return Y

def modified(Z,sx):
    # Ns = len(Z) - sx +1
    # Y = np.zeros(Ns)
    # for k in range(Ns):
    #     Y[k] = np.mean(Z[k:k+sx])
    
    Y = np.convolve(Z,np.ones(sx),'valid')/sx
    return Y 

def imf(Z,sx):
    Y = np.squeeze(np.sum(Z[:sx,:],axis=0))
    return Y

def timeshift(Z,sx):
    Y = np.transpose(np.reshape(Z[:sx*(len(Z)//sx)],(len(Z)//sx,sx)))
    return Y

def generalized(Z,sx):
    Ns = len(Z)//sx
    Y = np.var(np.reshape(Z[:sx*Ns],(Ns,sx)), axis=1)
    return Y
    

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


import numpy as np           
""" Base Approximate Entropy function."""

def ApEn(Sig, m=2, tau=1, r=None, Logx=np.exp(1)):
    """ApEn  estimates the approximate entropy of a univariate data sequence.
        
    .. code-block:: python
        
        Ap, Phi = ApEn(Sig)
    
    Returns the approximate entropy estimates (``Ap``) and the log-average number of 
    matched vectors (``Phi``) for ``m`` = [0,1,2], estimated for the data sequence (``Sig``)
    using the default parameters:
    embedding dimension = 2, time delay = 1, radius threshold = 0.2*SD(``Sig``), logarithm = natural
    
    .. code-block:: python
        
        Ap, Phi = ApEn(Sig, keyword = value, ...)
    
    Returns the approximate entropy estimates (``Ap``) of the data sequence (``Sig``)
    for dimensions = [0, 1, ..., ``m``] using the specified keyword arguments:
          :m:  - Embedding Dimension, a positive integer
          :tau: - Time Delay, a positive integer
          :r:  - Radius Distance Threshold, a positive scalar  
          :Logx: - Logarithm base, a positive scalar 
    
    :See also:
        ``XApEn``, ``SampEn``, ``MSEn``, ``FuzzEn``, ``PermEn``, ``CondEn``, ``DispEn``
    
    :References:
        [1] Steven M. Pincus, 
            "Approximate entropy as a measure of system complexity." 
            Proceedings of the National Academy of Sciences 
            88.6 (1991): 2297-2301.
    
    """
    
    Sig = np.squeeze(Sig)
    N = Sig.shape[0]
    if r is None:
        r = 0.2*np.std(Sig)
        
    assert N>10 and Sig.ndim == 1, "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and (r>=0), "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"

    Counter = 1*(abs(np.expand_dims(Sig,axis=0) -np.expand_dims(Sig,axis=1))<= r)  
    M = np.hstack((m*np.ones(N-m*tau), np.repeat(np.arange(m-1,0,-1),tau)))
    Ap = np.zeros((m+1))
    Phi = np.zeros((m+2))    
    for n in range(M.shape[0]):
        ix = np.where(Counter[n, :] == 1)[0]
        
        for k in range(1, int(M[n]+1)):
            ix = ix[ix + (k*tau) < N]     
            p1 = np.tile(Sig[n: n + 1 + (tau*k):tau], (ix.shape[0], 1))                       
            p2 = Sig[np.expand_dims(ix,axis=1) + np.arange(0,(k*tau)+1,tau)]           
            ix = ix[np.amax(abs(p1 - p2), axis=1) <= r] 
            Counter[n, ix] += 1
            
    #Phi[0] = (np.log(N)/np.log(Logx))/N
    Phi[1] = np.mean(np.log(np.sum(Counter>0,axis=0)/N)/np.log(Logx))
    Ap[0]  = Phi[0] - Phi[1]
    
    for k in range(m):
        ai = np.sum(Counter>k+1,axis=0)/(N-(k+1)*tau)
        bi = np.sum(Counter>k,axis=0)/(N-(k*tau))
        ai = ai[ai>0]
        bi = bi[bi>0]
        
        with np.errstate(divide='ignore', invalid='ignore'):
            Phi[k+2] = np.sum(np.log(ai)/np.log(Logx))/(N-(k+1)*tau)
            Ap[k+1]  = np.sum(np.log(bi)/np.log(Logx))/(N-(k*tau)) - Phi[k+2]
        
    return Ap, Phi

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


""" Base Multivariate Dispersion Entropy function."""
import numpy as np 
from scipy.special import ndtr
from scipy.cluster.vq import kmeans2 
from itertools import combinations as cmb # permutations as prm,
     

def MvDispEn(Data, m=None, tau=None, c=3, Typex='NCDF', Methodx='v1', Norm=False, Logx=np.exp(1)):
    """MvDispEn  estimates the multivariate dispersion entropy of a multivariate dataset.

    .. code-block:: python
    
        MDisp, RDE = MvDispEn(Data)
    
    Returns the multivariate dispersion entropy estimate (``MDisp``) and
    the reverse dispersion entropy (``RDE``) for the M multivariate sequences 
    in ``Data`` using the default parameters:
    embedding dimension = 2*np.ones(M,1), time delay = np.ones(M,1), # symbols = 3, 
    algorithm method = 'v1' (see below), data transform = normalised cumulative density function (ncdf)
    logarithm = natural, data normalization = true,
    
    
    .. important::

        By default, ``MvDispEn`` uses the method termed ``mvDEii`` in [1],
        which follows the original multivariate embedding approach of Ahmed & Mandic [2].
        The ``v1`` method therefore returns a singular entropy estimate.
           
        If the ``v2`` method is selected (``Methodx=='v2'``), the main method
        outlined in [1] termed ``mvDE`` is applied. In this case, entropy is estimated
        using each combination of multivariate delay vectors with lengths 1:max(m),
        with each entropy value returned accordingly. See [1] for more info.

 
    .. code-block:: python
    
        MDisp, RDE = MvDispEn(Data, keyword = value, ...)
        
    Returns the multivariate dispersion entropy estimate (``MDisp``) for the M
    multivariate data sequences in ``Data`` using the specified keyword arguments:
        :Data:     - Multivariate dataset, NxM matrix of N (>10) observations (rows) and M (cols) univariate data sequences
        :m:        - Embedding Dimension, a vector of M positive integers
        :tau:      - Time Delay, a vector of M positive integers
        :c:        - Number of symbols in transform, an integer > 1
        :Methodx:  - The method of multivariate dispersion entropy estimation as outlined in [1], either:
                :``'v1'``:                     employs the method consistent with the original multivariate 
                    embedding approach of Ahmed &  Mandic [2], termed ``mvDEii`` in [1]. (default)
                :``'v2'``:                      employs the main method derived in [1],  termed ``mvDE``.
        :Typex:    - Type of data-to-symbolic sequence transform, one of the following strings   {``'linear'``, ``'kmeans'``, ``'ncdf'``, ``'equal'``}
                     See the `EntropyHub Guide` for more info on these transforms.
        :Logx:  - Logarithm base, a positive scalar
        :Norm:  - Normalisation of ``MDisp`` and ``RDE`` values, a boolean:
                 :false:      no normalisation (default)
                 :true:       normalises w.r.t number of possible dispersion patterns (``c^m``).

    :See also:
        ``DispEn``, ``DispEn2D``, ``MvSampEn``, ``MvFuzzEn``, ``MvPermEn``, ``MSEn``
    
    :References:
      [1] H Azami, A Fernández, J Escudero
          "Multivariate Multiscale Dispersion Entropy of Biomedical Times Series"
          Entropy 2019, 21, 913.

      [2] Ahmed Mosabber Uddin, Danilo P. Mandic
          "Multivariate multiscale entropy: A tool for complexity
          analysis of multichannel data."
          Physical Review E 84.6 (2011): 061918.

      [3] Mostafa Rostaghi and Hamed Azami,
           "Dispersion entropy: A measure for time-series analysis." 
           IEEE Signal Processing Letters 
           23.5 (2016): 610-614.

      [4] Hamed Azami and Javier Escudero,
           "Amplitude-and fluctuation-based dispersion entropy." 
           Entropy 
           20.3 (2018): 210.

      [5] Li Yuxing, Xiang Gao and Long Wang,
           "Reverse dispersion entropy: A new complexity measure for sensor signal." 
           Sensors 
           19.23 (2019): 5203.
        
    """
   
    Data = np.squeeze(Data)
    assert Data.shape[0]>10 and Data.ndim==2 and Data.shape[1]>1,  "Data:   must be an NxM numpy matrix where N>10 and M>1"
    N, Dn = Data.shape 
    if m is None:    m = 2*np.ones(Dn, dtype=int)
    if tau is None:  tau = np.ones(Dn, dtype=int)
    m = m.astype(int)
    tau = tau.astype(int) 
  
    assert isinstance(m,np.ndarray) and all(m>0) and m.size==Dn and m.ndim==1, "m:     must be numpy vector of M positive integers"
    assert isinstance(tau,np.ndarray) and all(tau>0) and tau.size==Dn and tau.ndim==1, "tau:   must be numpy vector of M positive integers"
    assert isinstance(c,int) and (c > 1), "c:     must be an integer > 1"
    assert Typex.lower() in ['linear','kmeans','ncdf','equal'], \
                "Typex:    must be one of the following strings - 'linear', 'kmeans', 'ncdf', 'equal'"     
    assert isinstance(Methodx,str) and Methodx.lower() in ['v1','v2'], "Methodx:   must be either 'v1' or 'v2'"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - False or True"
       
    if Logx == 0: Logx = np.exp(1)
    
    Sx = np.zeros((N,Dn), dtype=np.int8)               
    for q in range(Dn):
        Sig = Data[:,q]        
        
        if Typex.lower() == 'linear':
            Zi = np.digitize(Sig,np.arange(min(Sig),max(Sig),np.ptp(Sig)/c))
            
        elif Typex.lower() == 'kmeans':        
            Clux, Zx = kmeans2(Sig, c, iter=200)
            Zx += 1;  xx = np.argsort(Clux) + 1;     Zi = np.zeros(N);
            for k in range(1,c+1):
                Zi[Zx==xx[k-1]] = k;
                
            del Zx, Clux, xx
            
        elif Typex.lower() == 'ncdf':  
            Zx = ndtr((Sig-np.mean(Sig))/np.std(Sig))
            Zi = np.digitize(Zx,np.arange(0,1,1/c))              
            del Zx
                             
        elif Typex.lower() == 'equal':
            ix = np.argsort(Sig, kind='mergesort')
            xx = np.round(np.arange(0,2*N,N/c)).astype(int)
            Zi = np.zeros(N)
            for k in range(c):
                Zi[ix[xx[k]:xx[k+1]]] = k+1 
            del ix, xx, k  
    
        Sx[:,q] = Zi
    
    Nx = N - max((m-1)*tau)
    Vex = np.zeros((Nx,sum(m)), dtype=np.int8)               
    q = 0 
    for k in range(Dn):
        for p in range(m[k]):
            Vex[:,q] = Sx[p*tau[k]:Nx+p*tau[k],  k]
            q += 1
        
                   
    if Methodx.lower() == 'v1':
        Px = np.unique(Vex,axis=0)
        Counter = np.zeros(Px.shape[0])
        for n in range(Px.shape[0]):
            Counter[n] = sum(np.any(Vex - Px[n,:],axis=1)==False)            
        Counter = Counter[Counter!=0]
        Ppi = Counter/sum(Counter)
        
        assert np.round(np.sum(Ppi),5) == 1, 'Warning: Potential error with probability calculation'
           
        with np.errstate(divide='ignore', invalid='ignore'):
            MDisp = -np.sum(Ppi*(np.log(Ppi)/np.log(Logx)))
            RDE = sum((Ppi - (1/(c**float(np.sum(m)))))**2)
            
        if Norm:
            MDisp = MDisp/(np.log(c**float(sum(m)))/np.log(Logx))
            RDE = RDE/(1-(1/(c**float(sum(m)))))

    
    elif Methodx.lower() == 'v2':
        P = sum(m)
        MDisp = np.zeros(max(m))
        RDE = np.zeros(max(m))
        for k in range(1,max(m)+1):
            print(' .', end='')
            Temp = np.array(list(cmb(np.arange(P),k)))
            Vez = np.zeros((Nx*nchk(P,k),k),dtype=np.int8)
            for q in range(Temp.shape[0]):
                Vez[q*Nx:(q+1)*Nx,:] = Vex[:,Temp[q]] 
            
            Px = np.unique(Vez,axis=0)
            Counter = np.zeros(Px.shape[0])
            for n in range(Px.shape[0]):
                Counter[n] = np.sum(np.any(Vez - Px[n,:],axis=1)==False)            
            Counter = Counter[Counter!=0]
            Ppi = Counter/np.sum(Counter)
        
            assert np.round(np.sum(Ppi),5) == 1, 'Warning: Potential error with probability calculation'
           
            with np.errstate(divide='ignore', invalid='ignore'):
                MDisp[k-1] = -np.sum(Ppi*(np.log(Ppi)/np.log(Logx)))
                RDE[k-1] = np.sum((Ppi - (1/(c**float(k))))**2)
        
            if Norm:
                MDisp[k-1] = MDisp[k-1]/(np.log(c**float(k))/np.log(Logx))
                RDE[k-1] = RDE[k-1]/(1-(1/(c**float(k))))
    
    return MDisp, RDE


def nchk(a,b):
    return int(np.math.factorial(a)/(np.math.factorial(b)*np.math.factorial(a-b)))

    
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


import numpy as np           
""" Base Range Entropy function."""

def RangEn(Sig, m=2, tau=1, r=0.2, Methodx='SampEn', Logx=np.exp(1)):
    """RangEn  estimates the range entropy of a univariate data sequence.

    .. code-block:: python
    
        Rangx, A, B = RangEn(Sig) 
        
    Returns the range entropy estimate (``Rangx``) and the number of matched state
    vectors (``m: B``, ``m+1: A``) estimated from the data sequence (``Sig``)
    using the sample entropy algorithm and the following default parameters: 
    embedding dimension = 2, time delay = 1, radius threshold = 0.2, logarithm = natural.    
    
    .. code-block:: python
    
        Rangx, A, B = RangEn(Sig, keyword = value, ...)
        
    Returns the range entropy estimates (``Rangx``) for dimensions = ``m``
    estimated for the data sequence (``Sig``) using the specified keyword arguments:
        :m:         - Embedding Dimension, a positive integer
        :tau:       - Time Delay, a positive integer
        :r:         - Radius Distance Threshold, a positive value between 0 and 1
        :Methodx:   - Base entropy method, either 'SampEn' [default] or 'ApEn'
        :Logx:      - Logarithm base, a positive scalar  
 
    :See also:
        ``ApEn``, ``SampEn``, ``FuzzEn``,  ``MSEn``
  
    :References:
        [1] Omidvarnia, Amir, et al.
            "Range entropy: A bridge between signal complexity and self-similarity"
            Entropy 
            20.12 (2018): 962.
            
        [2] Joshua S Richman and J. Randall Moorman. 
            "Physiological time-series analysis using approximate entropy
            and sample entropy." 
            American Journal of Physiology-Heart and Circulatory Physiology 
            2000

    """
  
    Sig = np.squeeze(Sig)
    N = Sig.shape[0]  
    
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and (r>=0) and (r<=1), "r:     must be a value between 0 and 1"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Methodx,str) and (Methodx.lower() in ['apen','sampen']), "Methodx must be either 'ApEn' or 'SampEn'"
            
    if Methodx.lower() == 'sampen':            
        Nx = N - m*tau
        Sx = np.zeros((Nx,m+1))
        for k in range(m+1):        
            Sx[:,k] = Sig[k*tau:Nx + k*tau]
       
        A = np.zeros(Nx,dtype=int)
        B = np.zeros(Nx,dtype=int)
        for k in range(Nx - 1):            
            Dxy = np.abs(np.tile(Sx[k,:-1],(Nx-(k+1),1)) - Sx[k+1:,:-1])             
            Mx = np.max(Dxy,axis=1)
            Mn = np.min(Dxy,axis=1)
            RR = (Mx - Mn)/(Mx + Mn) <= r
            #RR = Mx <= r
            B[k] = RR.sum()
            
            if B[k]>0:
                Dxy2 = np.abs(np.tile(Sx[k,:],(B[k],1)) - Sx[k+1:,:][RR,:])             
                Mx = np.max(Dxy2,axis=1)
                Mn = np.min(Dxy2,axis=1)
                RR2 = (Mx - Mn)/(Mx + Mn) <= r                            
                #RR2 = Mx <= r
                A[k] = RR2.sum()
                    
        with np.errstate(divide='ignore', invalid='ignore'):
            Rangx = -np.log(A.sum()/B.sum())/np.log(Logx) 
       
            return Rangx, A, B
        
    elif Methodx.lower() == 'apen':
        Nx = N - (m-1)*tau
        Sx = np.zeros((Nx,m))
        for k in range(m):        
            Sx[:,k] = Sig[k*tau:Nx + k*tau]                
        Sx = np.hstack((Sx, np.expand_dims(np.hstack((Sig[m*tau::],np.zeros(tau))), axis=1)))        

        B = np.zeros(Nx,dtype=int)
        A = np.zeros(Nx-tau,dtype=int)
        for k in range(Nx):                    
            Dxy = np.abs(np.tile(Sx[k,:-1],(Nx,1)) - Sx[:,:-1])             
            Mx = np.max(Dxy,axis=1)
            Mn = np.min(Dxy,axis=1)
            RR = (Mx - Mn)/(Mx + Mn) <= r             
            # RR = Mx <= r
            B[k] = RR.sum()
                     
            if k < (Nx - tau):
                RR[-tau:] = False
                Dxy2 = np.abs(np.tile(Sx[k,:],(RR.sum(),1)) - Sx[RR,:])             
                Mx2 = np.max(Dxy2,axis=1)
                Mn2 = np.min(Dxy2,axis=1)
                RR2 = (Mx2 - Mn2)/(Mx2 + Mn2) <= r                     
                # RR2 = Mx2 <= r                       
                A[k] = RR2.sum()
                        
        with np.errstate(divide='ignore', invalid='ignore'):
            Ax = (np.log(A/(Nx-tau))/np.log(Logx)).mean()
            Bx = (np.log(B/Nx)/np.log(Logx)).mean()
            
            Ap = Bx - Ax
       
            return Ap, A, B  
    
    else:
        raise Exception("Methodx must be either 'ApEn' or 'SampEn'")

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base Refined Cross-Multiscale Entropy function."
import numpy as np 
from scipy.signal import butter, filtfilt
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy

def rXMSEn(Sig1, Sig2, Mbjx, Scales=3, F_Order=6, F_Num=.5, RadNew=0, Plotx=False):
    """rXMSEn  returns the refined multiscale cross-entropy between two   univariate data sequences.
            
    .. code-block:: python

        MSx, CI = rXMSEn(Sig1, Sig2, Mobj) 
        
    Returns a vector of refined multiscale cross-entropy values (``MSx``) and
    the complexity index (``CI``) between the data sequences contained in ``Sig1`` and ``Sig2``
    using the parameters specified by the multiscale object (``Mobj``) and the
    following default parameters:   Scales = 3, Butterworth LPF Order = 6, 
    Butterworth LPF cutoff frequency at scale (T): Fc = 0.5/T. 
    If the entropy function specified by ``Mobj`` is ``XSampEn`` or ``XApEn``, ``rXMSEn``
    updates the threshold radius of the data sequences at each scale
    to 0.2*SDpooled(Sig1,Sig2) if no ``r`` value is provided by Mobj, or r*SDpooled(Sig1,Sig2) if ``r``
    is specified.
     
    .. code-block:: python
    
        MSx, CI = rXMSEn(Sig1, Sig2, Mobj, keyword = value, ...)
        
    Returns a vector of refined multiscale cross-entropy values (``MSx``) and 
    the complexity index (``CI``) between the data sequences contained in ``Sig1`` and ``Sig2`` 
    using the parameters specified by the multiscale object (``Mobj``) and the
    following 'keyword' arguments:
        :Scales:   - Number of temporal scales, an integer > 1 (default: 3)
        :F_Order:  - Butterworth low-pass filter order, a positive integer (default: 6)
        :F_Num:    - Numerator of Butterworth low-pass filter cutoff frequency, a scalar value in range [0 < ``F_Num`` < 1]. The cutoff frequency  at each scale (T) becomes: Fc = ``F_Num``/T.  (default: 0.5)
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the cross-entropy specified by ``Mobj`` is ``XSampEn`` or ``XApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Pooled Standard Deviation   - ``r*std(Ykj)``
                        * [2] Pooled Variance             - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                        
        :Plotx:   - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    :See also:
        ``MSobject``, ``XMSEn``, ``cXMSEn``, ``hXMSEn``, ``XSampEn``, ``XApEn``, ``MSEn``, ``rMSEn``
      
    :References:
        [1] Matthew W. Flood,
            "rXMSEn - EntropyHub Project"
            2024, https://github.com/MattWillFlood/EntropyHub
      
        [2] Rui Yan, Zhuo Yang, and Tao Zhang,
            "Multiscale cross entropy: a novel algorithm for analyzing two
            time series." 
            5th International Conference on Natural Computation. 
            Vol. 1, pp: 411-413 IEEE, 2009.
      
        [3] José Fernando Valencia, et al.,
            "Refined multiscale entropy: Application to 24-h holter 
            recordings of heart period variability in healthy and aortic 
            stenosis subjects." 
            IEEE Transactions on Biomedical Engineering 
            56.9 (2009): 2202-2213.
      
        [4] Puneeta Marwaha and Ramesh Kumar Sunkaria,
            "Optimal selection of threshold value ‘r’ for refined multiscale
            entropy." 
            Cardiovascular engineering and technology 
            6.4 (2015): 557-576.
      
        [5] Yi Yin, Pengjian Shang, and Guochen Feng, 
            "Modified multiscale cross-sample entropy for complex time 
            series."
            Applied Mathematics and Computation 
            289 (2016): 98-110.
      
        [6] Antoine Jamin, et al,
            "A novel multiscale cross-entropy method applied to navigation 
            data acquired with a bike simulator." 
            41st annual international conference of the IEEE EMBC
            IEEE, 2019.
      
        [7] Antoine Jamin and Anne Humeau-Heurtier. 
            "(Multiscale) Cross-Entropy Methods: A Review." 
            Entropy 
            22.1 (2020): 45.
      
    """
    
    Mobj = deepcopy(Mbjx)    
    S1 = np.squeeze(Sig1)
    S2 = np.squeeze(Sig2)
    assert Sig1.ndim==1 and Sig2.ndim==1,   "Sig1/Sig2:   must be numpy vectors (N>10)" 
    N1 = S1.shape[0]
    N2 = S2.shape[0]
                   
    assert N1>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"     
    assert isinstance(Mobj,object) and Mobj.Func.__name__[0]=='X', "Mobj:  must \
    be a x-multiscale entropy object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert isinstance(F_Order, int) and F_Order>0, "F_Order:  a positive integer"
    assert isinstance(F_Num,float) and 0<F_Num<1, "F_Num:  a scalar value in range [0 < F_Num < 1]"   
    assert (np.isin(RadNew,np.arange(1,5)) and Mobj.Func.__name__ in \
    ['XSampEn','XApEn']) or RadNew==0, "RadNew:     must be an integer in range \
    [1 4] and entropy function must be 'XSampEn' or 'XApEn'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
    
    if Mobj.Func.__name__ in 'XSampEn': Mobj.Kwargs['Vcp'] = False 
    
    MSx = np.zeros(Scales)
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x,y: np.sqrt((np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1))
        elif RadNew == 2:
            Rnew = lambda x,y: (np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1)
        elif RadNew == 3:
            Rnew = lambda x,y: np.mean(abs(np.hstack((x,y))-np.mean(np.hstack((x,y)))))
        elif RadNew == 4:
            Rnew = lambda x,y: np.median(abs(np.hstack((x,y))-np.median(np.hstack((x,y)))))   
                
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Pooled Standard Deviation','Pooled Variance','Joint Mean Abs Deviation', 'Joint Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])              
            Cx = .2

    for T in range(1,Scales+1):
        print(' .', end='')
        Temp = refined(S1,S2,T,F_Order,F_Num)
        if RadNew > 0:
            Mobj.Kwargs.update({'r': Cx*Rnew(Temp[0],Temp[1])})            
        Tempx = Mobj.Func(Temp[0],Temp[1],**Mobj.Kwargs)
        if isinstance(Tempx,tuple):
            if isinstance(Tempx[0],(int,float)):
                MSx[T-1] = Tempx[0]
            else:
                MSx[T-1] =Tempx[0][-1]
        elif isinstance(Tempx,(int,float)):
                MSx[T-1] = Tempx
        elif isinstance(Tempx,np.ndarray):
                MSx[T-1] = Tempx[-1]
            
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')
    
    print('\n')

    if Plotx:       
       figure()
       ax1 = axes()   
       ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
       ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
       ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_title('Refined Multiscale %s'%(Mobj.Func.__name__), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))      
       show()
    
    return MSx, CI


def refined(Za, Zb, sx, P1, P2):
    bb, aa = butter(P1, P2/sx)
    Ya = filtfilt(bb, aa, Za)
    Yb = filtfilt(bb, aa, Zb)
    T1 = Ya[::sx]
    T2 = Yb[::sx]    
    return T1, T2

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Cross Distribution Entropy function."""
import numpy as np    
from scipy.stats import skew

def XDistEn(*Sig, m=2, tau=1, Bins='Sturges', Logx=2, Norm=True):
    """XDistEn  estimates the cross-distribution entropy between two univariate data sequences.
    
    .. code-block:: python
    
        XDist, Ppi = XDistEn(Sig1, Sig2) 
        
    Returns the cross-distribution entropy estimate (``XDist``) and the
    corresponding distribution probabilities (``Ppi``) estimated between the data 
    sequences contained in ``Sig1`` and ``Sig2`` using the default parameters: 
    embedding dimension = 2, time delay = 1, binning method = ``'Sturges'``,
    logarithm = base 2, normalisation = w.r.t # of histogram bins
 
    .. code-block:: python
    
        XDist, Ppi = XDistEn(Sig1, Sig2, keyword = value, ...)
        
    Returns the cross-distribution entropy estimate (``XDist``) estimated between the 
    data sequences contained in ``Sig1`` and ``Sig2`` using the specified 'keyword' = arguments:
        :m:     - Embedding Dimension, a positive integer   [default: 2]
        :tau:   - Time Delay, a positive integer            [default: 1]
        :Bins:  - Histogram bin selection method for distance distribution,
            * an integer > 1 indicating the number of bins, or one of the 
            * following strings {``'sturges'``, ``'sqrt'``, ``'rice'``, ``'doanes'``} [default: 'sturges']
        :Logx:  - Logarithm base, a positive scalar [default: 2] (enter 0 for natural log)
        :Norm:  - Normalisation of DistEn value, a boolean value:
            * [False]  no normalisation.
            * [True]   normalises w.r.t # of histogram bins [default]
    
    :See also: 
        ``XSampEn``, ``XApEn``, ``XPermEn``, ``XCondEn``, ``DistEn``, ``DistEn2D``, ``XMSEn``
    
    :References:
        [1] Yuanyuan Wang and Pengjian Shang,
            "Analysis of financial stock markets through the multiscale
            cross-distribution entropy based on the Tsallis entropy."
            Nonlinear Dynamics 
            94.2 (2018): 1361-1376.
    
    """
    
    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])
        
    N  = S1.shape[0]
    N2 = S2.shape[0]
    
    if Logx == 0:
        Logx = np.exp(1)
    assert N>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert isinstance(m,int) and (m>0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau>0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"
    assert (isinstance(Bins,int) and Bins > 1) \
            or (Bins.lower() in ['sturges','sqrt','rice','doanes']), \
            "Bins:    an integer > 1, or can be one of the following strings - \
            'sturges', 'sqrt', 'rice', 'doanes'"    
       
    Nx = N - ((m-1)*tau) 
    Nx2 = N2 - ((m-1)*tau) 
    Zm1 = np.zeros((Nx,m))
    Zm2 = np.zeros((Nx2,m))
    for n in np.arange(m):
        Zm1[:,n] = S1[n*tau:Nx+(n*tau)]
        Zm2[:,n] = S2[n*tau:Nx2+(n*tau)]
        
    DistMat = np.zeros((Nx,Nx2))
    for k in range(Nx):
        DistMat[k:] = np.max(abs(np.tile(Zm1[k,:],(Nx2,1))-Zm2),axis=1)
    
    # Nx = len(S1) - ((m-1)*tau)   
    # Zm1 = np.zeros((Nx,m))
    # Zm2 = np.zeros((Nx,m))
    # for n in np.arange(m):
    #     Zm1[:,n] = S1[n*tau:Nx+(n*tau)]
    #     Zm2[:,n] = S2[n*tau:Nx+(n*tau)]
        
    # DistMat = np.zeros((Nx,Nx))
    # for k in range(Nx):
    #     DistMat[k:] = np.max(abs(np.tile(Zm1[k,:],(Nx,1))-Zm2),axis=1)
      
    Ny = Nx*Nx2
    DistMat = np.reshape(DistMat,(Ny))  
    if isinstance(Bins, str):
        if Bins.lower() == 'sturges':
            Bx = np.ceil(np.log2(Ny) + 1)
        elif Bins.lower() == 'rice':
            Bx = np.ceil(2*(Ny**(1/3)))
        elif Bins.lower() == 'sqrt':
            Bx = np.ceil(np.sqrt(Ny))
        elif Bins.lower()== 'doanes':
            sigma = np.sqrt(6*(Ny-2)/((Ny+1)*(Ny+3)))
            Bx = np.ceil(1+np.log2(Ny)+np.log2(1+abs(skew(DistMat)/sigma)))
        else:
            raise Exception('Please enter a valid binning method')               
    else:
        Bx = Bins
        
    Ppi,_ = np.histogram(DistMat,int(Bx))        
    Ppi = Ppi/Ny    
    if round(sum(Ppi),6) != 1:
        print('Warning: Potential error estimating probabilities (p = ' +str(np.sum(Ppi))+ '.')
        Ppi = Ppi[Ppi!=0]
    elif any(Ppi==0):
        print('Note: '+str(sum(Ppi==0))+'/'+str(len(Ppi))+' bins were empty')
        Ppi = Ppi[Ppi!=0]
           
    XDist = -sum(Ppi*np.log(Ppi)/np.log(Logx))
    if Norm:
        XDist = XDist/(np.log(Bx)/np.log(Logx))
      
    return XDist, Ppi

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base cross-Multiscale Entropy function."
import numpy as np 
from PyEMD import EMD
from matplotlib.pyplot import figure, axes,show
from copy import deepcopy

def XMSEn(Sig1, Sig2, Mbjx, Scales=3, Methodx='coarse', RadNew=0, Plotx=False):
    """XMSEn  returns the multiscale cross-entropy between two univariate data sequences.
 
    .. code-block:: python
    
        MSx, CI = XMSEn(Sig1, Sig2, Mobj) 
        
    Returns a vector of multiscale cross-entropy values (``MSx``) and the
    complexity index (``CI``) between the data sequences contained in ``Sig1`` and ``Sig2`` using 
    the parameters specified by the multiscale object (``Mobj``) over 3 temporal
    scales with coarse-graining (default).
     
    .. code-block:: python
    
        MSx, CI = XMSEn(Sig1, Sig2, Mobj, keyword = value, ...)
        
    Returns a vector of multiscale cross-entropy values (``MSx``) and the
    complexity index (``CI``) between the data sequences contained in ``Sig1`` and ``Sig2`` 
    using the parameters specified by the multiscale object (``Mobj``) and the
    following 'keyword' arguments:
        :Scales:   - Number of temporal scales, an integer > 1  [default: 3]
        :Methodx:  - Graining method, one of the following: [default: ``'coarse'``]  {``'coarse'``, ``'modified'``, ``'imf'`` , ``'timeshift'``, ``'generalized'``}
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the cross-entropy specified by ``Mobj`` is ``XSampEn`` or ``XApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Pooled Standard Deviation   - ``r*std(Ykj)``
                        * [2] Pooled Variance             - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                     
        :Plotx:    - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve)  [default: False]
    
    :See also:
        ``MSobject``, ``XSampEn``, ``XApEn``, ``rXMSEn``, ``cXMSEn``, ``hXMSEn``, ``MSEn``
    
    :References:
        [1] Rui Yan, Zhuo Yang, and Tao Zhang,
            "Multiscale cross entropy: a novel algorithm for analyzing two
            time series." 
            5th International Conference on Natural Computation. 
            Vol. 1, pp: 411-413 IEEE, 2009.
    
        [2] Madalena Costa, Ary Goldberger, and C-K. Peng,
            "Multiscale entropy analysis of complex physiologic time series."
            Physical review letters
            89.6 (2002): 068102.
    
        [3] Vadim V. Nikulin, and Tom Brismar,
            "Comment on “Multiscale entropy analysis of complex physiologic
            time series”." 
            Physical review letters 
            92.8 (2004): 089803.
    
        [4] Madalena Costa, Ary L. Goldberger, and C-K. Peng. 
            "Costa, Goldberger, and Peng reply." 
            Physical Review Letters
            92.8 (2004): 089804.
    
        [5] Antoine Jamin, et al,
            "A novel multiscale cross-entropy method applied to navigation 
            data acquired with a bike simulator." 
            41st annual international conference of the IEEE EMBC
            IEEE, 2019.
    
        [6] Antoine Jamin and Anne Humeau-Heurtier. 
            "(Multiscale) Cross-Entropy Methods: A Review." 
            Entropy 
            22.1 (2020): 45.
    
    """
    
    Mobj = deepcopy(Mbjx)    
    S1 = np.squeeze(Sig1)
    S2 = np.squeeze(Sig2)
    assert Sig1.ndim==1 and Sig2.ndim==1,   "Sig1/Sig2:   must be numpy vectors (N>10)" 
    N1 = S1.shape[0]
    N2 = S2.shape[0]

    Chk = ['coarse','modified','imf','timeshift','generalized']  
    assert N1>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"  
    assert isinstance(Mobj,object) and Mobj.Func.__name__[0]=='X', "Mobj:  must \
    be a x-multiscale entropy object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert Methodx.lower() in Chk, "Methodx:  must be one of the following string names- \
    'coarse', 'modified' , 'imf', 'timeshift', 'generalized'"    
    assert isinstance(RadNew,int) and (np.isin(RadNew,np.arange(1,5)) \
                and Mobj.Func.__name__ in ['XSampEn','XApEn']) or RadNew==0, \
    "RadNew:     must be 0, or an integer in range [1 4] with entropy function 'XSampEn' or 'XApEn'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
            
    if Mobj.Func.__name__ in 'XSampEn': Mobj.Kwargs['Vcp'] = False 
    
    if Methodx.lower()=='imf':
        EMD().FIXE = 100
        EMD().FIXE_H = 100
        S1 = EMD().emd(S1,max_imf=Scales-1).T    
        S2 = EMD().emd(S2,max_imf=Scales-1).T   

    Func2 = globals()[Methodx.lower()]
    MSx = np.zeros(Scales)
        
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x,y: np.sqrt((np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1))
        elif RadNew == 2:
            Rnew = lambda x,y: (np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1)
        elif RadNew == 3:
            Rnew = lambda x,y: np.mean(abs(np.hstack((x,y))-np.mean(np.hstack((x,y)))))
        elif RadNew == 4:
            Rnew = lambda x,y: np.median(abs(np.hstack((x,y))-np.median(np.hstack((x,y)))))   
        
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Pooled Standard Deviation','Pooled Variance','Joint Mean Abs Deviation', 'Joint Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])            
            Cx = .2

    for T in range(1,Scales+1):
        print(' .', end='')
        Temp = Func2(S1,S2,T)  
        
        if Methodx.lower() == 'timeshift':
            Tempx = np.zeros(T)
            for k in range(T):
                print(' .', end='')
                if RadNew > 0:
                    Mobj.Kwargs.update({'r': Cx*Rnew(Temp[0][k,:],Temp[1][k,:])})                                    
                Tempy = Mobj.Func(Temp[0][k,:],Temp[1][k,:],**Mobj.Kwargs)
                
                if isinstance(Tempy,tuple):
                    if isinstance(Tempy[0],(int,float)):
                        Tempx[k] = Tempy[0]
                    else:
                        Tempx[k] =Tempy[0][-1]
                elif isinstance(Tempy,(int,float)):
                    Tempx[k] = Tempy
                elif isinstance(Tempy,np.ndarray):
                    Tempx[k] = Tempy[-1]   
                    
            Temp2 = np.mean(Tempx)
            
        else:
            if RadNew > 0:
                    Mobj.Kwargs.update({'r': Cx*Rnew(Temp[0],Temp[1])})            
            Tempx = Mobj.Func(Temp[0],Temp[1],**Mobj.Kwargs)
            
            if isinstance(Tempx,tuple):
                if isinstance(Tempx[0],(int,float)):
                    Temp2 = Tempx[0]
                else:
                    Temp2 =Tempx[0][-1]
            elif isinstance(Tempx,(int,float)):
                Temp2 = Tempx
            elif isinstance(Tempx,np.ndarray):
                Temp2 = Tempx[-1]        
        MSx[T-1] = Temp2
            
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')
    
    print('\n')
    if Plotx:
       figure()
       ax1 = axes()            
       ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
       ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
       ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_title('Cross-Multiscale %s (%s-graining method)'%(Mobj.Func.__name__,Methodx), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))      
       show()
    
    return MSx, CI

def coarse(Za,Zb,sx):
    Na = len(Za)//sx
    Nb = len(Zb)//sx
    T1 = np.mean(np.reshape(Za[:sx*Na],(Na,sx)),axis=1)
    T2 = np.mean(np.reshape(Zb[:sx*Nb],(Nb,sx)),axis=1)
    return T1, T2

def modified(Za,Zb,sx):    
    T1 = np.convolve(Za,np.ones(sx),'valid')/sx
    T2 = np.convolve(Zb,np.ones(sx),'valid')/sx
    return T1, T2

def timeshift(Za,Zb,sx):
    # Y = np.zeros((sx,len(Z)//sx,2))
    # Y[:,:,0] = np.reshape(Z[:sx*(len(Z)//sx),0],(len(Z)//sx,sx)).transpose()
    # Y[:,:,1] = np.reshape(Z[:sx*(len(Z)//sx),1],(len(Z)//sx,sx)).transpose()   
    
    T1 = np.zeros((sx,len(Za)//sx))
    T2 = np.zeros((sx,len(Zb)//sx))
    T1 = np.reshape(Za[:sx*(len(Za)//sx)],(len(Za)//sx,sx)).T
    T2 = np.reshape(Zb[:sx*(len(Zb)//sx)],(len(Zb)//sx,sx)).T   
    return T1, T2

def generalized(Za,Zb,sx):
    Na = len(Za)//sx
    Nb = len(Zb)//sx
    T1 = np.var(np.reshape(Za[:sx*Na],(Na,sx)),axis=1)
    T2 = np.var(np.reshape(Zb[:sx*Nb],(Nb,sx)),axis=1)
    return T1, T2

def imf(Za,Zb,sx):
    # Y = np.squeeze(np.sum(Z[:,:,:sx],axis=2))    
    T1 = np.squeeze(np.sum(Za[:,:sx],axis=1))
    T2 = np.squeeze(np.sum(Zb[:,:sx],axis=1))
    return T1, T2

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""